"""Agents that use precomputed values in the protocol."""

from sail_on_client.agent.visual_agent import VisualAgent
from sail_on_client.agent.ond_agent import ONDAgent
from sail_on_client.agent.condda_agent import CONDDAAgent
from typing import Dict, Any, Tuple, Callable

import logging
import os
import pandas as pd

log = logging.getLogger(__name__)


class PreComputedAgent(VisualAgent):
    """Detector for submitting precomputed results."""

    def __init__(
        self,
        algorithm_name: str,
        cache_dir: str,
        has_roundwise_file: bool,
        round_size: int,
    ) -> None:
        """
        Construct agent with precomputed results.

        Args:
            algorithm_name: Name of the algorithm
            cache_dir: Path to cache directory
            has_roundwise_file: Flag to determine if the cache has files for rounds
            round_size: Size of a round
        """
        ONDAgent.__init__(self)
        self.algorithm_name = algorithm_name
        self.cache_dir = cache_dir
        self.has_roundwise_file = has_roundwise_file
        self.round_size = round_size
        self.step_dict: Dict[str, Callable] = {
            "Initialize": self.initialize,
            "FeatureExtraction": self.feature_extraction,
            "WorldDetection": self.world_detection,
        }

    def get_config(self) -> Dict:
        """Return a default configuration dictionary."""
        return {
            "algorithm_name": self.algorithm_name,
            "cache_dir": self.cache_dir,
            "has_roundwise_file": self.has_roundwise_file,
            "round_size": self.round_size,
        }

    def execute(self, toolset: Dict, step_descriptor: str) -> Any:
        """
        Execute method used by the protocol to run different steps associated with the algorithm.

        Args:
            toolset (dict): Dictionary containing parameters for different steps
            step_descriptor (str): Name of the step
        """
        log.info(f"Executing {step_descriptor}")
        return self.step_dict[step_descriptor](toolset)

    def initialize(self, toolset: Dict) -> None:
        """
        Algorithm Initialization.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            None
        """
        self.round_idx = {"detection": 0, "classification": 0}
        self.test_id = toolset["test_id"]

    def _get_test_info_from_toolset(self, toolset: Dict) -> Tuple:
        """
        Private function for getting test id and round id (optionally) from toolset.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            tuple containing test id and round id (optionally)
        """
        if self.has_roundwise_file:
            return (self.test_id, toolset["round_id"])
        else:
            return self.test_id

    def _get_result_path(self, toolset: Dict, step_descriptor: str) -> str:
        """
        Private function for getting path to results.

        Args:
            toolset (dict): Dictionary containing parameters for different steps
            step_descriptor (str): Name of the step

        Return:
            Path to the result file
        """
        if self.has_roundwise_file:
            test_id, round_id = self._get_test_info_from_toolset(toolset)
            return os.path.join(
                self.cache_dir,
                f"{test_id}.{round_id}_{self.algorithm_name}_{step_descriptor}.csv",
            )
        else:
            test_id = self._get_test_info_from_toolset(toolset)
            return os.path.join(
                self.cache_dir, f"{test_id}_{self.algorithm_name}_{step_descriptor}.csv"
            )

    def _generate_step_result(self, toolset: Dict, step_descriptor: str) -> str:
        result_path = self._get_result_path(toolset, step_descriptor)
        if self.has_roundwise_file:
            return result_path
        else:
            round_file_path = os.path.join(
                self.cache_dir, f"{self.algorithm_name}_{step_descriptor}.csv"
            )
            round_idx = self.round_idx[step_descriptor]
            test_df = pd.read_csv(result_path, header=None)
            round_df = test_df.iloc[round_idx : round_idx + self.round_size]
            self.round_idx[step_descriptor] += self.round_size
            round_df.to_csv(round_file_path, index=False, header=False)
            return round_file_path

    def feature_extraction(
        self, toolset: Dict
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Feature extraction step for the algorithm.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            Tuple of dictionary
        """
        self.dataset = toolset["dataset"]
        pd.read_csv(self.dataset, header=None).shape[0]
        return {}, {}

    def world_detection(self, toolset: Dict) -> str:
        """
        Detect change in world ( Novelty has been introduced ).

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            path to csv file containing the results for change in world
        """
        return self._generate_step_result(toolset, "detection")


class PreComputedONDAgent(PreComputedAgent, ONDAgent):
    """Detector for submitting precomputed results in OND."""

    def __init__(
        self,
        algorithm_name: str,
        cache_dir: str,
        has_roundwise_file: bool,
        round_size: int,
    ) -> None:
        """
        Construct agent with precomputed results for OND.

        Args:
            algorithm_name: Name of the algorithm
            cache_dir: Path to cache directory
            has_roundwise_file: Flag to determine if the cache has files for rounds
            round_size: Size of a round
        """
        PreComputedAgent.__init__(
            self, algorithm_name, cache_dir, has_roundwise_file, round_size
        )
        ONDAgent.__init__(self)
        self.step_dict.update(
            {
                "NoveltyClassification": self.novelty_classification,
                "NoveltyAdaptation": self.novelty_adaptation,
                "NoveltyCharacterization": self.novelty_characterization,
            }
        )

    def novelty_classification(self, toolset: Dict) -> str:
        """
        Classify data provided in known classes and unknown class.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            path to csv file containing the results for novelty classification step
        """
        return self._generate_step_result(toolset, "classification")

    def novelty_adaptation(self, toolset: Dict) -> None:
        """
        Update models based on novelty classification and characterization.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            None
        """
        pass

    def novelty_characterization(self, toolset: Dict) -> str:
        """
        Characterize novelty by clustering different novel samples.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            path to csv file containing the results for novelty characterization step
        """
        return self._get_result_path(toolset, "characterization")


class PreComputedCONDDAAgent(PreComputedAgent, CONDDAAgent):
    """Detector for submitting precomputed results in CONDDA."""

    def __init__(
        self,
        algorithm_name: str,
        cache_dir: str,
        has_roundwise_file: bool,
        round_size: int,
    ) -> None:
        """
        Construct agent with precomputed results for CONDDA.

        Args:
            algorithm_name: Name of the algorithm
            cache_dir: Path to cache directory
            has_roundwise_file: Flag to determine if the cache has files for rounds
            round_size: Size of a round
        """
        PreComputedAgent.__init__(
            self, algorithm_name, cache_dir, has_roundwise_file, round_size
        )
        ONDAgent.__init__(self)
        self.step_dict.update(
            {"NoveltyCharacterization": self.novelty_characterization}
        )

    def novelty_characterization(self, toolset: Dict) -> str:
        """
        Characterize novelty by clustering different novel samples.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            path to csv file containing the results for novelty characterization step
        """
        return self._get_result_path(toolset, "characterization")



"""Constants used by the file provider."""


class ProtocolConstants:
    """Class with constants for the protocol."""

    # Feedback types
    DETECTION = "detection"
    CLASSIFICATION = "classification"
    CHARACTERIZATION = "characterization"
    TRANSCRIPTION = "transcription"
    TEMPORAL = "temporal"
    SPATIAL = "spatial"
    PSUEDO_CLASSIFICATION = "psuedo_labels_classification"
    SCORE = "score"

    # Detection requirement
    REQUIRED = "REQUIRED"
    NOTIFY_AND_CONTINUE = "NOTIFY_AND_CONTINUE"
    IGNORE = "IGNORE"
    SKIP = "SKIP"



"""Document Transcription Feedback."""

import pandas as pd
from sail_on_client.harness.local_harness import LocalHarness
from sail_on_client.harness.par_harness import ParHarness
from sail_on_client.feedback.feedback import Feedback

from typing import Union, Dict

SUPPORTED_FEEDBACK = ["classification", "score", "transcription"]


class DocumentTranscriptionFeedback(Feedback):
    """Feedback for document transcription."""

    def __init__(
        self,
        first_budget: int,
        income_per_batch: int,
        maximum_budget: int,
        interface: Union[LocalHarness, ParHarness],
        session_id: str,
        test_id: str,
        feedback_type: str,
    ) -> None:
        """
        Initialize document transcription feedback object.

        Args:
            first_budget: Initial budget
            income_per_batch: Additional labels added after every batch
            maximum_budget: Max labels that can be requested
            interface: An instance of evaluation interface
            session_id: Session identifier
            test_id: Test identifier
            feedback_type: Type of feedback that can be requested

        Returns:
            None
        """
        if feedback_type not in SUPPORTED_FEEDBACK:
            raise ValueError(f"Unsupported feedback_type {feedback_type}")
        super(DocumentTranscriptionFeedback, self).__init__(
            first_budget,
            income_per_batch,
            maximum_budget,
            interface,
            session_id,
            test_id,
            feedback_type,
        )
        self.current_round: int = -1
        self.budget: int = first_budget

    def get_levenshtein_feedback(
        self, round_id: int, images_id_list: list, image_names: list
    ) -> Union[Dict, None]:
        """
        Get levenshtein feedback for the round.

        Args:
            round_id: Round identifier
            image_id_list: List if indices for images
            image_names: List of image names for the round

        Return:
            A dictionary containing levenshtein score or None if
            feedback is requested for an older round
        """
        if round_id > self.current_round:
            self.deposit_income()
            self.current_round = round_id
            if len(images_id_list) <= self.budget:
                self.budget = self.budget - len(images_id_list)
                image_ids = [image_names[int(idx)] for idx in images_id_list]
                feedback_file = self.interface.get_feedback_request(
                    image_ids,
                    self.feedback_type,
                    self.test_id,
                    round_id,
                    self.session_id,
                )
                df = pd.read_csv(feedback_file, delimiter=",", header=None)
                return df
            else:
                raise ValueError("the function should be added")
        else:
            return None

    def get_feedback(
        self, round_id: int, images_id_list: list, image_names: list
    ) -> Union[pd.DataFrame, Dict, None]:
        """
        Get feedback for the round.

        Args:
            round_id: Round identifier
            image_id_list: List if indices for images
            image_names: List of image names for the round

        Return:
            Either a dataframe or dictionary with score if the request is valid
            for the current round.
        """
        if self.feedback_type == "classification":
            feedback_fn = self.get_labeled_feedback
        elif self.feedback_type == "score":
            feedback_fn = self.get_score_feedback
        elif self.feedback_type == "transcription":
            feedback_fn = self.get_levenshtein_feedback
        else:
            raise ValueError("Unsupported feedback type {self.feedback_type} specified")
        return feedback_fn(round_id, images_id_list, image_names)



"""Implementation of T&E Harness for running experiments locally."""

from sail_on_client.harness.file_provider import FileProvider
from sail_on_client.harness.file_provider_fn import get_session_info
from sail_on_client.errors import RoundError as ClientRoundError
from sail_on_client.evaluate import create_metric_instance
from sail_on_client.harness.test_and_evaluation_harness import TestAndEvaluationHarness

from tempfile import TemporaryDirectory
from typing import Any, Dict, Union, List
import os
import io
import logging
import ubelt as ub
import pandas as pd
import json

log = logging.getLogger(__name__)


class LocalHarness(TestAndEvaluationHarness):
    """Harness without any server communication."""

    def __init__(
        self, data_dir: str, result_dir: str, gt_dir: str = "", gt_config: str = ""
    ) -> None:
        """
        Initialize an object of local harness.

        Args:
            data_dir: Path to the directory with the data
            result_dir: Path to the directory where results are stored
            gt_dir: Path to directory with ground truth
            gt_config: Path to config file with column mapping for ground truth

        Returns:
            None
        """
        TestAndEvaluationHarness.__init__(self)
        self.temp_dir = TemporaryDirectory()
        self.data_dir = data_dir
        self.gt_dir = gt_dir
        self.gt_config = gt_config
        self.temp_dir_name = self.temp_dir.name
        self.result_dir = result_dir
        self.file_provider = FileProvider(self.data_dir, self.result_dir)

    def get_config(self) -> Dict:
        """JSON Compliant representation of the object."""
        return {
            "data_dir": self.data_dir,
            "gt_dir": self.gt_dir,
            "result_dir": self.result_dir,
            "gt_config": self.gt_config,
        }

    def test_ids_request(
        self,
        protocol: str,
        domain: str,
        detector_seed: str,
        test_assumptions: str = "{}",
    ) -> str:
        """
        Request Test Identifiers as part of a series of individual tests.

        Args:
            protocol : string indicating which protocol is being evaluated
            domain : problem domain for the tests
            detector_seed : A seed provided by the novelty detector
            test_assumptions : Assumptions used by the detector

        Returns:
            Filename of file containing test ids
        """
        result_dict = self.file_provider.test_ids_request(
            protocol, domain, detector_seed, test_assumptions
        )
        return result_dict["test_ids"]

    def session_request(
        self,
        test_ids: list,
        protocol: str,
        domain: str,
        novelty_detector_version: str,
        hints: list,
        detection_threshold: float,
    ) -> str:
        """
        Create a new session to evaluate the detector using an empirical protocol.

        Args:
            test_ids: List of tests being evaluated in this session
            protocol: String indicating which protocol is being evaluated
            domain: String indicating which domain is being evaluated
            novelty_detector_version: The novelty detector being evaluated
            hints: Hints used for the session
            detection_threshold: Detection threshold for the session

        Returns:
            A session identifier provided by the server
        """
        return self.file_provider.new_session(
            test_ids,
            protocol,
            domain,
            novelty_detector_version,
            hints,
            detection_threshold,
        )

    def resume_session(self, session_id: str) -> List[str]:
        """
        Get finished test from an existing session.

        Args:
            session id : Session id that was started but not terminated

        Returns:
            List of tests finished in the session
        """
        return self.file_provider.latest_session_info(session_id)["finished_tests"]

    def dataset_request(self, test_id: str, round_id: int, session_id: str) -> str:
        """
        Request data for evaluation.

        Args:
            test_id: The test being evaluated at this moment.
            round_id: The sequential number of the round being evaluated
            session_id: The identifier provided by the server for a single experiment

        Returns:
            Filename of a file containing a list of image files (including full path for each)
        """
        self.data_file = os.path.join(
            self.temp_dir_name, f"{session_id}.{test_id}.{round_id}.csv"
        )
        byte_stream = self.file_provider.dataset_request(session_id, test_id, round_id)
        if byte_stream is None:
            raise ClientRoundError(
                reason="End of Dataset", msg="All Data from dataset has been requested"
            )
        else:
            with open(self.data_file, "wb") as f:
                f.write(byte_stream.getbuffer())
            return self.data_file

    def get_feedback_request(
        self,
        feedback_ids: list,
        feedback_type: str,
        test_id: str,
        round_id: int,
        session_id: str,
    ) -> str:
        """
        Get Labels from the server based provided one or more example ids.

        Args:
            feedback_ids: List of media ids for which feedback is required
            feedback_type: Protocols constants with the values: label, detection, characterization
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            Path to a file containing containing requested feedback
        """
        self.feedback_file = os.path.join(
            self.temp_dir_name,
            "feedback",
            f"{session_id}.{test_id}.{round_id}_{feedback_type}.csv",
        )
        ub.ensuredir(os.path.join(self.temp_dir_name, "feedback"))
        byte_stream = self.file_provider.get_feedback(
            feedback_ids, feedback_type, session_id, test_id
        )
        with open(self.feedback_file, "wb") as f:
            f.write(byte_stream.getbuffer())
        return self.feedback_file

    def post_results(
        self, result_files: Dict[str, str], test_id: str, round_id: int, session_id: str
    ) -> None:
        """
        Post client detector predictions for the dataset.

        Args:
            result_files: A dictionary of results with protocol constant as key and file path as value
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            None
        """
        info = get_session_info(str(self.result_dir), session_id)
        protocol = info["created"]["protocol"]
        domain = info["created"]["domain"]
        base_result_path = os.path.join(str(self.result_dir), protocol, domain)
        os.makedirs(base_result_path, exist_ok=True)
        result_content = {}
        for result_key, result_file in result_files.items():
            with open(result_file, "r") as f:
                content = f.read()
            result_content[result_key] = io.StringIO(content).getvalue()
        self.file_provider.post_results(session_id, test_id, round_id, result_content)

    def evaluate_round_wise(
        self, test_id: str, round_id: int, session_id: str,
    ) -> Dict[str, Any]:
        """
        Get results for round(s).

        Args:
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            Path to a file with the results
        """
        gt_file_id = os.path.join(self.gt_dir, f"{test_id}_single_df.csv")
        gt = pd.read_csv(gt_file_id, sep=",", header=None, skiprows=1, quotechar="|")
        info = get_session_info(str(self.result_dir), session_id)
        protocol = info["created"]["protocol"]
        domain = info["created"]["domain"]
        metadata = self.get_test_metadata(session_id, test_id)
        round_size = metadata["round_size"]
        results: Dict[str, Union[Dict, float]] = {}
        gt_config = json.load(open(self.gt_config, "r"))
        classification_file_id = os.path.join(
            self.result_dir,
            protocol,
            domain,
            f"{session_id}.{test_id}_classification.csv",
        )

        classifications = pd.read_csv(
            classification_file_id, sep=",", header=None, quotechar="|"
        )
        gt_round = gt.iloc[round_id * round_size : (round_id + 1) * round_size]
        classification_round = classifications[
            round_id * round_size : (round_id + 1) * round_size
        ]
        metric = create_metric_instance(protocol, domain, gt_config)
        m_acc = metric.m_acc_round_wise(
            classification_round, gt_round[metric.classification_id], round_id
        )
        results[f"m_acc_round_{round_id}"] = m_acc
        log.info(f"Accuracy for {test_id}, {round_id}: {ub.repr2(results)}")
        return results

    def evaluate(
        self,
        test_id: str,
        round_id: int,
        session_id: str,
        baseline_session_id: str = None,
    ) -> Dict[str, Any]:
        """
        Get results for test(s).

        Args:
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            Path to a file with the results
        """
        gt_file_id = os.path.join(self.gt_dir, f"{test_id}_single_df.csv")
        gt = pd.read_csv(gt_file_id, sep=",", header=None, skiprows=1, quotechar="|")
        info = get_session_info(str(self.result_dir), session_id)
        protocol = info["created"]["protocol"]
        domain = info["created"]["domain"]
        results: Dict[str, Union[Dict, float]] = {}
        gt_config = json.load(open(self.gt_config, "r"))
        detection_file_id = os.path.join(
            self.result_dir, protocol, domain, f"{session_id}.{test_id}_detection.csv",
        )
        detections = pd.read_csv(detection_file_id, sep=",", header=None, quotechar="|")
        classification_file_id = os.path.join(
            self.result_dir,
            protocol,
            domain,
            f"{session_id}.{test_id}_classification.csv",
        )

        classifications = pd.read_csv(
            classification_file_id, sep=",", header=None, quotechar="|"
        )
        if baseline_session_id is not None:
            baseline_classification_file_id = os.path.join(
                self.result_dir,
                protocol,
                domain,
                f"{baseline_session_id}.{test_id}_classification.csv",
            )
            baseline_classifications = pd.read_csv(
                baseline_classification_file_id, sep=",", header=None
            )
        metric = create_metric_instance(protocol, domain, gt_config)
        detection_idx = 1
        gt_classification_idx = metric.classification_id
        # ######## Image Classification Evaluation  ###########
        if domain == "image_classification":
            gt_detection_idx = metric.detection_id
            # p_unknown column for image classification
            novel_idx = 1
        # ######## Activity Recognition Evaluation  ###########
        elif domain == "activity_recognition":
            detection_idx = 1
            gt_detection_idx = metric.novel_id
            # p_unknown column for activity recognition
            novel_idx = 31
        # ######## Document Transcription Evaluation  ###########
        elif domain == "transcripts":
            detection_idx = 1
            gt_detection_idx = metric.novel_id
            # p_unknown column for transcripts in writer identification
            novel_idx = 50
        else:
            raise AttributeError(
                f'Domain: "{domain}" is not a real domain.  Get a clue.'
            )

        m_num = metric.m_num(detections[detection_idx], gt[gt_detection_idx])
        results["m_num"] = m_num
        m_num_stats = metric.m_num_stats(
            detections[detection_idx], gt[gt_detection_idx]
        )
        results["m_num_stats"] = m_num_stats
        m_ndp = metric.m_ndp(classifications[novel_idx], gt[gt_detection_idx])
        results["m_ndp"] = m_ndp
        m_ndp_pre = metric.m_ndp_pre(classifications[novel_idx], gt[gt_detection_idx])
        results["m_ndp_pre"] = m_ndp_pre
        m_ndp_post = metric.m_ndp_post(classifications[novel_idx], gt[gt_detection_idx])
        results["m_ndp_post"] = m_ndp_post
        m_acc = metric.m_acc(
            gt[gt_detection_idx], classifications, gt[gt_classification_idx], 100, 5,
        )
        results["m_acc"] = m_acc
        m_acc_failed = metric.m_ndp_failed_reaction(
            detections[detection_idx],
            gt[gt_detection_idx],
            classifications,
            gt[gt_classification_idx],
        )
        results["m_acc_failed"] = m_acc_failed
        m_is_cdt_and_is_early = metric.m_is_cdt_and_is_early(
            m_num_stats["GT_indx"], m_num_stats["P_indx_0.5"], gt.shape[0],
        )
        results["m_is_cdt_and_is_early"] = m_is_cdt_and_is_early
        if baseline_session_id is not None:
            m_acc_baseline = metric.m_acc(
                gt[gt_detection_idx],
                baseline_classifications,
                gt[gt_classification_idx],
                100,
                5,
            )
            log.info(f"Baseline performance for {test_id}: {ub.repr2(m_acc_baseline)}")
            m_nrp = metric.m_nrp(m_acc, m_acc_baseline)
            results["m_nrp"] = m_nrp

        log.info(f"Results for {test_id}: {ub.repr2(results)}")
        return results

    def get_test_metadata(self, session_id: str, test_id: str) -> Dict[str, Any]:
        """
        Retrieve the metadata json for the specified test.

        Args:
            session_id: The id of the session currently being evaluated
            test_id: The id of the test currently being evaluated

        Returns:
            A dictionary containing metadata
        """
        return self.file_provider.get_test_metadata(session_id, test_id)

    def complete_test(self, session_id: str, test_id: str) -> None:
        """
        Mark test as completed.

        Args:
            session_id: The id of the session currently being evaluated
            test_id: The id of the test currently being evaluated

        Returns:
            None
        """
        self.file_provider.complete_test(session_id, test_id)

    def terminate_session(self, session_id: str) -> None:
        """
        Terminate the session after the evaluation for the protocol is complete.

        Args:
            session_id: The id provided by a server denoting a session

        Returns: None
        """
        self.file_provider.terminate_session(session_id)



"""Checkpoint to save and restore attributes."""

import logging
import torch
import os
import pickle as pkl
from torch import Tensor

from typing import Dict, Any

log = logging.getLogger(__name__)


class Checkpointer(object):
    """Checkpoint object to save and restore attributes."""

    def __init__(self, toolset: Dict) -> None:
        """
        Initialize.

        Args:
            toolset: Dictionary with parameter for the mixin

        Returns:
            None
        """
        self.toolset = toolset

    def _save_elementwise_attribute(
        self, detector: Any, attribute: str, attribute_dict: Dict
    ) -> Dict:
        """
        Private method to save attributes element wise.

        Args:
            detector: Instance of novelty detector
            attribute: Name of the detector attribute that needs to be saved
            attribute dict: A dictonary containing attribute value for other tests

        Returns:
            Update attribute dictionary
        """
        attribute_val = getattr(detector, attribute)
        test_id = self.toolset["test_id"]
        if isinstance(attribute_val, dict):
            if test_id in attribute_dict:
                attribute_dict[test_id].update(attribute_val)
            else:
                attribute_dict[test_id] = attribute_val
        elif isinstance(attribute_val, list):
            if test_id in attribute_dict:
                attribute_dict[test_id].extend(attribute_val)
            else:
                attribute_dict[test_id] = attribute_val
        elif isinstance(attribute_val, tuple):
            if test_id in attribute_dict:
                old_attr_val = list(attribute_dict[test_id])
                old_attr_val.extend(list(attribute_val))
                attribute_dict[test_id] = tuple(old_attr_val)
            else:
                attribute_dict[test_id] = attribute_val
        elif isinstance(attribute_val, Tensor):
            if test_id in attribute_dict:
                attribute_dict[test_id] = torch.cat(
                    [attribute_dict[test_id], attribute_val]
                )
            else:
                attribute_dict[test_id] = attribute_val
        else:
            log.info(
                "Treating attribute value as a single element rather than an iterable"
            )
            attribute_dict[test_id] = attribute_val
        return attribute_dict

    def save_attributes(self, step_descriptor: str) -> None:
        """
        Save attribute for a detector.

        Args:
            step_descriptor: String describing steps for protocol

        Returns
            None
        """
        if step_descriptor in self.toolset["saved_attributes"]:
            attributes = self.toolset["saved_attributes"][step_descriptor]
        else:
            attributes = []
        save_elementwise = self.toolset["save_elementwise"]
        attribute_dict = self.toolset["attributes"]
        self.detector: Any
        if len(attributes) > 0:
            for attribute in attributes:
                if hasattr(self.detector, attribute) and save_elementwise:
                    if attribute not in attribute_dict:
                        attribute_dict[attribute] = {}
                    attribute_dict[attribute] = self._save_elementwise_attribute(
                        self.detector, attribute, attribute_dict[attribute]
                    )
                elif hasattr(self.detector, attribute) and not save_elementwise:
                    raise NotImplementedError(
                        "Saving attributes for an entire round is not supported"
                    )
                else:
                    log.warn(f"Detector does not have {attribute} attribute")
        else:
            log.info(f"No attributes found for {step_descriptor}")
        self.toolset["attributes"] = attribute_dict

    def _restore_elementwise_attribute(
        self, detector: Any, attribute_name: str, attribute_val: Dict
    ) -> Any:
        """
        Private method to restore attributes element wise.

        Args:
            detector: Instance of novelty detector
            attribute_name: Name of the detector attribute that needs to be saved
            attribute_val: A dictonary containing value for attributes

        Returns
            Detector with updated value for attributes
        """
        dataset_ids = list(
            map(lambda x: x.strip(), open(self.toolset["dataset"], "r").readlines())
        )
        round_id = self.toolset["round_id"]
        round_len = len(dataset_ids)
        if isinstance(attribute_val, dict):
            round_attribute_val = {}
            for dataset_id in dataset_ids:
                round_attribute_val[dataset_id] = attribute_val[dataset_id]
        elif (
            isinstance(attribute_val, list)
            or isinstance(attribute_val, tuple)
            or isinstance(attribute_val, Tensor)
        ):
            round_attribute_val = attribute_val[
                round_id * round_len : (round_id + 1) * round_len
            ]
        else:
            log.info(
                "Treating attribute value as a single element rather than an iterable."
            )
            round_attribute_val = attribute_val
        setattr(detector, attribute_name, round_attribute_val)
        return detector

    def restore_attributes(self, step_descriptor: str) -> None:
        """
        Restore attribute for a detector.

        Args:
            step_descriptor: String describing steps for protocol

        Returns:
            None
        """
        if (
            self.toolset["use_saved_attributes"]
            and step_descriptor in self.toolset["saved_attributes"]
        ):
            attributes = self.toolset["saved_attributes"][step_descriptor]
            save_elementwise = self.toolset["save_elementwise"]
            save_dir = self.toolset["save_dir"]
            test_id = self.toolset["test_id"]
            if os.path.isdir(save_dir):
                attribute_file = os.path.join(save_dir, f"{test_id}_attribute.pkl")
                attribute_val = pkl.load(open(attribute_file, "rb"))
            else:
                attribute_val = pkl.load(open(save_dir, "rb"))
            for attribute in attributes:
                if save_elementwise:
                    self.detector = self._restore_elementwise_attribute(
                        self.detector, attribute, attribute_val[attribute][test_id]
                    )
                else:
                    raise NotImplementedError(
                        "Restoring attributes for an entire round is not supported."
                    )

        else:
            log.info(f"No attributes found for {step_descriptor}.")



"""Round for visual protocol."""

import logging
from typing import List, Any, Tuple, Dict

from sail_on_client.protocol.visual_dataclasses import (
    FeatureExtractionParams,
    WorldChangeDetectionParams,
)
from sail_on_client.harness.test_and_evaluation_harness import (
    TestAndEvaluationHarnessType,
)
from sail_on_client.utils.utils import safe_remove
from sail_on_client.utils.decorators import skip_stage


log = logging.getLogger(__name__)


class VisualRound:
    """Class with common elements for visual protocols."""

    def __init__(
        self,
        algorithm: Any,
        data_root: str,
        features_dict: Dict,
        harness: TestAndEvaluationHarnessType,
        logit_dict: Dict,
        redlight_instance: str,
        session_id: str,
        skip_stages: List[str],
        test_id: str,
    ) -> None:
        """
        Construct VisualRound.

        Args:
            algorithm: An instance of algorithm
            data_root: Root directory of the data
            features_dict: Dictionary with features for the entire dataset
            harness: An instance of the harness used for T&E
            logit_dict: Dictionary with logits for the entire dataset
            redlight_instance: The instance when the world changes
            session_id: Session id associated with the algorithm
            skip_stages: List of stages that are skipped
            test_id: Test id associated with the round

        Returns:
            None
        """
        self.algorithm = algorithm
        self.data_root = data_root
        self.features_dict = features_dict
        self.harness = harness
        self.logit_dict = logit_dict
        self.redlight_instance = redlight_instance
        self.session_id = session_id
        self.skip_stages = skip_stages
        self.test_id = test_id

    @staticmethod
    def get_instance_ids(dataset_path: str) -> List[str]:
        """
        Get instance ids from the dataset.

        Args:
            dataset_path: Path to text file with instances used in a round

        Returns:
            List of instance ids from the dataset
        """
        with open(dataset_path, "r") as dataset:
            instance_ids = dataset.readlines()
            instance_ids = [instance_id.strip() for instance_id in instance_ids]
        return instance_ids

    @skip_stage("FeatureExtraction", ({}, {}))
    def _run_feature_extraction(
        self, fe_params: FeatureExtractionParams, instance_ids: List[str]
    ) -> Tuple[Dict, Dict]:
        """
        Private helper function for running feature extraction.

        Args:
            fe_params: An instance of dataclass with parameters for feature extraction
            instance_ids: Identifiers associated with data for a round

        Returns:
            Tuple for feature and logit dictionary for a round
        """
        rfeature_dict, rlogit_dict = {}, {}
        if len(self.features_dict) > 0 and len(self.logit_dict) > 0:
            for instance_id in instance_ids:
                rfeature_dict[instance_id] = self.features_dict[instance_id]
                rlogit_dict[instance_id] = self.logit_dict[instance_id]
        else:
            fe_toolset = fe_params.get_toolset()
            rfeature_dict, rlogit_dict = self.algorithm.execute(
                fe_toolset, "FeatureExtraction"
            )
        self.rfeature_dict, self.rlogit_dict = rfeature_dict, rlogit_dict
        return rfeature_dict, rlogit_dict

    @skip_stage("WorldDetection")
    def _run_world_change_detection(
        self, wcd_params: WorldChangeDetectionParams, round_id: int,
    ) -> None:
        """
        Private helper function for detecting that the world has changed.

        Args:
            wcd_params: An instance of dataclass with parameters for world change detection
            round_id: Identifier for a round

        Returns:
            None
        """
        wd_result = self.algorithm.execute(wcd_params.get_toolset(), "WorldDetection")
        self.harness.post_results(
            {"detection": wd_result}, self.test_id, round_id, self.session_id
        )
        safe_remove(wd_result)



"""Abstract interface for an agent that works under OND."""

from sail_on_client.agent.visual_agent import VisualAgent
from typing import Dict
from abc import abstractmethod


class ONDAgent(VisualAgent):
    """Abstract class for OND agent."""

    def get_config(self) -> Dict:
        """Return a default configuration dictionary."""
        return {}

    @abstractmethod
    def novelty_classification(self, ncl_toolset: Dict) -> str:
        """
        Abstract method for novelty classification.

        Args:
            ncl_toolset: Parameters for feature extraction

        Returns:
            Path to results for novelty classification.
        """
        pass

    @abstractmethod
    def novelty_adaptation(self, na_toolset: Dict) -> None:
        """
        Abstract method for novelty adaptation.

        Args:
            na_toolset: Parameters for novelty adaptation

        Returns:
            None
        """
        pass

    @abstractmethod
    def novelty_characterization(self, nc_toolset: Dict) -> str:
        """
        Abstract method for novelty novelty_characterization.

        Args:
            nc_toolset: Parameters for feature extraction

        Returns:
            Path to results for novelty characterization.
        """
        pass



"""Decorators for sail-on-client."""

import functools
from typing import Any, AnyStr, Callable


def skip_stage(stage_name: str, skip_return: Any = None) -> Callable:
    """
    Skip stages in the protocol.

    Args:
        stage_name: Name of the stage that is covered by the decorated function
        skip_stages: List of stages that should be skipped
        skip_return: Optional return types when the stage is skipped

    Returns:
        Decorated function call
    """

    def skip_stage_decorator(stage_fn: Callable) -> Callable:
        """
        Capture the stage function.

        Args:
            stage_fn: The callable function that would be wrapped

        Returns:
            Wrapped function
        """

        @functools.wraps(stage_fn)
        def skip_stage_fn(self: Any, *args: AnyStr, **kwargs: AnyStr) -> Any:
            if hasattr(self, "skip_stages"):
                skip_stages = self.skip_stages
            else:
                raise ValueError("The class does not skip_stages")

            if stage_name in skip_stages:
                return skip_return
            else:
                return stage_fn(self, *args, **kwargs)

        return skip_stage_fn

    return skip_stage_decorator



"""Activity Recognition Feedback."""

from sail_on_client.harness.local_harness import LocalHarness
from sail_on_client.harness.par_harness import ParHarness
from sail_on_client.feedback.feedback import Feedback
from typing import Union, Optional, Dict
import pandas as pd

SUPPORTED_FEEDBACK = [
    "classification",
    "score",
    "detection",
    "detection_and_classification",
]


class ActivityRecognitionFeedback(Feedback):
    """Feedback for activity recognition."""

    def __init__(
        self,
        first_budget: int,
        income_per_batch: int,
        maximum_budget: int,
        interface: Union[LocalHarness, ParHarness],
        session_id: str,
        test_id: str,
        feedback_type: str,
    ) -> None:
        """
        Initialize activity recognition feedback object.

        Args:
            first_budget: Initial budget
            income_per_batch: Additional labels added after every batch
            maximum_budget: Max labels that can be requested
            interface: An instance of evaluation interface
            session_id: Session identifier
            test_id: Test identifier
            feedback_type: Type of feedback that can be requested

        Returns:
            None
        """
        if feedback_type not in SUPPORTED_FEEDBACK:
            raise ValueError(f"Unsupported feedback_type {feedback_type}")
        super(ActivityRecognitionFeedback, self).__init__(
            first_budget,
            income_per_batch,
            maximum_budget,
            interface,
            session_id,
            test_id,
            feedback_type,
        )
        self.current_round: int = -1
        self.budget: int = first_budget

    def get_labeled_feedback(
        self, round_id: int, images_id_list: list, image_names: list
    ) -> Union[pd.DataFrame, None]:
        """
        Get labeled feedback for the round.

        Args:
            round_id: Round identifier
            image_id_list: List if indices for images
            image_names: List of image names for the round

        Return:
            A dictionary with the accuracy value or None if
            feedback is requested for an older round
        """
        if round_id > self.current_round:
            self.deposit_income()
            self.current_round = round_id
            if len(images_id_list) <= self.budget:
                self.budget = self.budget - len(images_id_list)
                image_ids = [image_names[int(idx)] for idx in images_id_list]
                feedback_file = self.interface.get_feedback_request(
                    image_ids,
                    self.feedback_type,
                    self.test_id,
                    round_id,
                    self.session_id,
                )
                df = pd.read_csv(
                    feedback_file,
                    delimiter=",",
                    header=None,
                    names=["id", "class1", "class2", "class3", "class4", "class5"],
                )
            else:
                raise ValueError("the function should be added")
        else:
            df = None
        return df

    def get_detection_feedback(
        self, round_id: int, images_id_list: list, image_names: list
    ) -> Optional[pd.DataFrame]:
        """
        Get detection feedback for the round.

        Args:
            round_id: Round identifier
            image_id_list: List of indices for images
            image_names: List of image names for the round

        Return:
            A dataframe with id and novelty detection value as columns
        """
        if round_id > self.current_round:
            self.deposit_income()
            self.current_round = round_id
            if len(images_id_list) <= self.budget:
                self.budget = self.budget - len(images_id_list)
                image_ids = [image_names[int(idx)] for idx in images_id_list]
                feedback_file = self.interface.get_feedback_request(
                    image_ids,
                    self.feedback_type,
                    self.test_id,
                    round_id,
                    self.session_id,
                )
                df = pd.read_csv(
                    feedback_file,
                    delimiter=",",
                    header=None,
                    names=["id", "detection"],
                )
            else:
                raise ValueError("Requesting over budget")
        else:
            df = None
        return df

    def get_detection_and_classification_feedback(
        self, round_id: int, images_id_list: list, image_names: list
    ) -> Optional[pd.DataFrame]:
        """
        Get detection and classification feedback for the round.

        Args:
            round_id: Round identifier
            image_id_list: List of indices for images
            image_names: List of image names for the round

        Return:
            A dataframe with id and novelty detection value as columns
        """
        if round_id > self.current_round:
            self.deposit_income()
            self.current_round = round_id
            if len(images_id_list) <= self.budget:
                self.budget = self.budget - len(images_id_list)
                image_ids = [image_names[int(idx)] for idx in images_id_list]
                detection_feedback_file = self.interface.get_feedback_request(
                    image_ids, "detection", self.test_id, round_id, self.session_id,
                )
                detection_df = pd.read_csv(
                    detection_feedback_file,
                    delimiter=",",
                    header=None,
                    names=["id", "detection"],
                )
                classification_feedback_file = self.interface.get_feedback_request(
                    image_ids,
                    "classification",
                    self.test_id,
                    round_id,
                    self.session_id,
                )
                classification_df = pd.read_csv(
                    classification_feedback_file,
                    delimiter=",",
                    header=None,
                    names=["id", "class1", "class2", "class3", "class4", "class5"],
                )
                df = pd.merge(detection_df, classification_df, on="id")
            else:
                raise ValueError("Requesting over budget")
        else:
            df = None
        return df

    def get_feedback(
        self, round_id: int, images_id_list: list, image_names: list
    ) -> Union[pd.DataFrame, Dict, None]:
        """
        Get feedback for the round.

        Args:
            round_id: Round identifier
            image_id_list: List if indices for images
            image_names: List of image names for the round

        Return:
            Either a dataframe or dictionary with score if the request is valid
            for the current round.
        """
        if self.feedback_type == "classification" or self.feedback_type == "score":
            return super().get_feedback(round_id, images_id_list, image_names)
        elif self.feedback_type == "detection":
            return self.get_detection_feedback(round_id, images_id_list, image_names)
        elif self.feedback_type == "detection_and_classification":
            return self.get_detection_and_classification_feedback(
                round_id, images_id_list, image_names
            )
        else:
            raise ValueError("Unsupported feedback type {self.feedback_type} specified")



"""Abstract Class for metrics for sail-on."""

import numpy as np
from abc import ABC, abstractmethod
from typing import Dict


class ProgramMetrics(ABC):
    """Abstract program metric class."""

    def __init__(self, protocol: str) -> None:
        """
        Initialize.

        Args:
            protocol: Name of the protocol.

        Returns:
            None
        """
        self.protocol = protocol

    @abstractmethod
    def m_acc(
        self,
        gt_novel: np.ndarray,
        p_class: np.ndarray,
        gt_class: np.ndarray,
        round_size: int,
        asymptotic_start_round: int,
    ) -> Dict:
        """
        m_acc abstract function.

        Args:
            gt_novel: ground truth detections
            p_class: class predictions
            gt_class: ground truth classes
            round_size: size of the round
            asymptotic_start_round: asymptotic samples considered for computing metrics

        Returns:
            Dictionary containing top1, top3 accuracy over the test, pre and post novelty.
        """

    @abstractmethod
    def m_acc_round_wise(
        self, p_class: np.ndarray, gt_class: np.ndarray, round_id: int
    ) -> Dict:
        """
        m_acc_round_wise abstract function.

        Args:
            p_class: class predictions
            gt_class: ground truth classes

        Returns:
            Dictionary containing top1, top3 accuracy for a round
        """

    @abstractmethod
    def m_num(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        m_num abstract function.

        Args:
            p_novel: detection predictions
            gt_novel: ground truth detections

        Returns:
            Difference between the novelty introduction and predicting change in world.
        """

    @abstractmethod
    def m_num_stats(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        m_num_stats abstract function.

        Args:
            p_novel: detection predictions
            gt_novel: ground truth detections

        Returns:
            Dictionary containing indices for novelty introduction and change in world prediction.
        """

    @abstractmethod
    def m_ndp(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        m_ndp abstract function.

        Args:
            p_novel: detection predictions
            gt_novel: ground truth detections

        Returns:
            Dictionary containing novelty detection performance over the test.
        """

    @abstractmethod
    def m_ndp_pre(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        m_ndp_pre abstract function.

        Args:
            p_novel: detection predictions
            gt_novel: ground truth detections

        Returns:
            Dictionary containing detection performance pre novelty.
        """

    @abstractmethod
    def m_ndp_post(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        m_ndp_post abstract function.

        Args:
            p_novel: detection predictions
            gt_novel: ground truth detections

        Returns:
            Dictionary containing detection performance post novelty.
        """

    @abstractmethod
    def m_ndp_failed_reaction(
        self,
        p_novel: np.ndarray,
        gt_novel: np.ndarray,
        p_class: np.ndarray,
        gt_class: np.ndarray,
    ) -> Dict:
        """
        m_ndp_failed_reaction abstract function.

        Args:
            p_novel: detection predictions
            gt_novel: ground truth detections
            p_class: class predictions
            gt_class: ground truth classes

        Returns:
            Dictionary containing TP, FP, TN, FN, top1, top3 accuracy over the test.
        """

    @abstractmethod
    def m_accuracy_on_novel(
        self, p_novel: np.ndarray, gt_class: np.ndarray, gt_novel: np.ndarray
    ) -> Dict:
        """
        m_accuracy_on_novel abstract function.

        Args:
            p_novel: detection predictions
            gt_class: ground truth classes
            gt_novel: ground truth detections

        Returns:
            Accuracy on novely samples
        """

    @abstractmethod
    def m_is_cdt_and_is_early(self, gt_idx: int, ta2_idx: int, test_len: int) -> Dict:
        """
        m_is_cdt_and_is_early abstract function.

        Args:
            gt_idx: Index when novelty is introduced
            ta2_idx: Index when change is detected
            test_len: Length of test

        Returns
            Dictionary containing boolean showing if change was was detected and if it was detected early
        """



"""A provider class for evaluating with files."""

from sail_on_client.errors.errors import ProtocolError, ServerError, RoundError
from sail_on_client.harness.constants import ProtocolConstants
from sail_on_client.harness.file_provider_fn import (
    get_session_info,
    get_session_test_info,
    log_session,
    read_meta_data,
    read_gt_csv_file,
    get_classification_feedback,
    get_classificaton_score_feedback,
    get_detection_feedback,
    get_classification_var_feedback,
    get_levenshtein_feedback,
    psuedo_label_feedback,
    write_session_log_file,
)

import csv
import logging
import os
import glob
import uuid
import traceback
from typing import List, Dict, Any, Optional
from io import BytesIO


class FileProvider:
    """File-based service provider."""

    def __init__(self, folder: str, results_folder: str) -> None:
        """
        Initialize file provider.

        Args:
            folder: Folder where the data is present
            results_folder: Folder where the results are saved

        Returns:
            None
        """
        self.folder = folder
        self.results_folder = results_folder
        os.makedirs(results_folder, exist_ok=True)

    def get_test_metadata(
        self,
        session_id: str,
        test_id: str,
        api_call: bool = True,
        in_process_only: bool = True,
    ) -> Dict[str, Any]:
        """
        Get test metadata.

        Args:
            session_id: Session id for which the info is required
            test_id: Test id for which the info is required
            api_call: Flag to change metadata to approved metadata
            in_process_only: Flag to get information while the session is active

        Returns:
            Metadata associated with the test as a dictionary
        """
        try:
            structure = get_session_info(
                self.results_folder, session_id, in_process_only=in_process_only
            )
            info = structure["created"]
            metadata_location = os.path.join(
                self.folder,
                info["protocol"],
                info["domain"],
                f"{test_id}_metadata.json",
            )
        except KeyError:
            raise ProtocolError(
                "session_id_invalid",
                f"Provided session id {session_id} could not be found or was improperly set up",
            )

        if not os.path.exists(metadata_location):
            raise ServerError(
                "metadata_not_found",
                f"Metadata file for Test Id {test_id} could not be found",
                "".join(traceback.format_stack()),
            )

        hints = []

        # List of metadata vars approved to be sent to the client
        approved_metadata = [
            "protocol",
            "known_classes",
            "max_novel_classes",
            "round_size",
            "feedback_max_ids",
            "pre_novelty_batches",
            "max_detection_feedback_ids",
        ]

        hints = info.get("hints", [])

        approved_metadata.extend([data for data in ["red_light"] if data in hints])

        overrides = {}
        for hint_data in hints:
            if "=" in hint_data:
                parts = hint_data.split("=")
                if parts[0] in approved_metadata:
                    overrides[parts[0]] = int(parts[1])

        md = read_meta_data(metadata_location)
        md.update(overrides)
        if api_call:
            return {k: v for k, v in md.items() if k in approved_metadata}
        return md

    def test_ids_request(
        self, protocol: str, domain: str, detector_seed: str, test_assumptions: str
    ) -> Dict[str, str]:
        """
        Request test IDs.

        Args:
            protocol: Name of the protocol
            domain: Domain of the application
            detector_seed: Seed used by the detector
            test_assumptions: Assumptions made when generating the tests

        Returns:
            Dictionary with the list of tests and seed used for generating tests
        """

        def _strip_id(filename: str) -> str:
            return os.path.splitext(os.path.basename(filename))[0]

        file_location = os.path.join(self.folder, protocol, domain, "test_ids.csv")
        if not os.path.exists(file_location):
            if not os.path.exists(os.path.join(self.folder, protocol)):
                msg = f"{protocol} not configured"
            elif not os.path.exists(os.path.join(self.folder, protocol, domain)):
                msg = f"domain {domain} for {protocol} not configured"
            else:
                test_ids = [
                    _strip_id(f)
                    for f in glob.glob(
                        os.path.join(self.folder, protocol, domain, "*.csv")
                    )
                ]
                with open(file_location, "w") as f:
                    f.writelines(test_ids)
                return {"test_ids": file_location, "generator_seed": "1234"}
            raise ProtocolError(
                "BadDomain", msg, "".join(traceback.format_stack()),
            )

        return {"test_ids": file_location, "generator_seed": "1234"}

    def new_session(
        self,
        test_ids: List[str],
        protocol: str,
        domain: str,
        novelty_detector_version: str,
        hints: List[str],
        detection_threshold: float,
    ) -> str:
        """
        Create a session.

        Args:
            test_ids: Test ids evaluated in the session
            protocol: Name of the protocol
            domain: Domain of the protocol
            novelty_detector_version: Version of the novelty detector
            hints: List of hints used in the session
            detection_threshold: Detection threshold used by the novelty detector

        Returns:
            Session id for the agent
        """
        # Verify's that all given test id's are valid and have associated csv files
        for test_id in test_ids:
            file_location = os.path.join(
                self.folder, protocol, domain, f"{test_id}_single_df.csv"
            )
            if not os.path.exists(file_location):
                raise ServerError(
                    "test_id_invalid",
                    f"Test Id {test_id} could not be matched to a specific file",
                    "".join(traceback.format_stack()),
                )

        session_id = str(uuid.uuid4())

        log_session(
            self.results_folder,
            session_id,
            activity="created",
            content={
                "protocol": protocol,
                "domain": domain,
                "detector": novelty_detector_version,
                "detection_threshold": detection_threshold,
                "hints": hints,
            },
        )

        return session_id

    def dataset_request(
        self, session_id: str, test_id: str, round_id: int
    ) -> Optional[BytesIO]:
        """
        Request a dataset.

        Args:
            test_id: The test being evaluated at this moment.
            round_id: The sequential number of the round being evaluated
            session_id: The identifier provided by the server for a single experiment

        Returns:
            Returns an instance of BytesIO
        """
        try:
            info = get_session_info(self.results_folder, session_id)["created"]
            test_info = get_session_test_info(self.results_folder, session_id, test_id)
            file_location = os.path.join(
                self.folder,
                info["protocol"],
                info["domain"],
                f"{test_id}_single_df.csv",
            )
        except KeyError:
            raise ProtocolError(
                "session_id_invalid",
                f"Provided session id {session_id} could not be found or was improperly set up",
            )

        if not os.path.exists(file_location):
            raise ServerError(
                "test_id_invalid",
                f"Test Id {test_id} could not be matched to a specific file",
                "".join(traceback.format_stack()),
            )

        metadata = self.get_test_metadata(session_id, test_id, False)

        if round_id is not None:
            # Check for removing leftover files from restarting tests within a session
            if int(round_id) == 0 and test_info:
                test_session_path = os.path.join(
                    self.results_folder, f"{str(session_id)}.{str(test_id)}.json"
                )
                if os.path.exists(test_session_path):
                    os.remove(test_session_path)
                test_result_paths = glob.glob(
                    os.path.join(
                        self.results_folder,
                        info["protocol"],
                        info["domain"],
                        f"{str(session_id)}.{str(test_id)}_*.csv",
                    )
                )
                for path in test_result_paths:
                    os.remove(path)

            temp_file_path = BytesIO()
            lines = read_gt_csv_file(file_location)
            lines = [x[0] for x in lines if x[0].strip("\n\t\"',.") != ""]
            try:
                round_pos = int(round_id) * int(metadata["round_size"])
            except KeyError:
                raise RoundError(
                    "no_defined_rounds",
                    f"round_size not defined in metadata for test id {test_id}",
                    "".join(traceback.format_stack()),
                )
            if round_pos >= len(lines):
                return None

            text = (
                "\n".join(lines[round_pos : round_pos + int(metadata["round_size"])])
                + "\n"
            ).encode("utf-8")
            temp_file_path.write(text)
            temp_file_path.seek(0)
        else:
            temp_file_path = open(file_location, "rb")

        log_session(
            self.results_folder,
            session_id,
            test_id=test_id,
            round_id=round_id,
            activity="data_request",
        )

        return temp_file_path

    feedback_request_mapping: Dict = {
        "image_classification": {
            ProtocolConstants.CLASSIFICATION: {
                "function": get_classification_feedback,
                "files": [ProtocolConstants.CLASSIFICATION],
                "columns": [1],
                "detection_req": ProtocolConstants.NOTIFY_AND_CONTINUE,
                "budgeted_feedback": True,
            },
            ProtocolConstants.SCORE: {
                "function": get_classificaton_score_feedback,
                "files": [ProtocolConstants.CLASSIFICATION],
                "columns": [1],
                "detection_req": ProtocolConstants.SKIP,
                "budgeted_feedback": False,
            },
            ProtocolConstants.DETECTION: {
                "function": get_detection_feedback,
                "files": [ProtocolConstants.DETECTION],
                "columns": [0],
                "detection_req": ProtocolConstants.SKIP,
                "budgeted_feedback": True,
                "required_hints": [],
                "alternate_budget": "max_detection_feedback_ids",
            },
        },
        "transcripts": {
            ProtocolConstants.CLASSIFICATION: {
                "function": get_classification_feedback,
                "files": [ProtocolConstants.CLASSIFICATION],
                "columns": [4],
                "detection_req": ProtocolConstants.SKIP,
                "budgeted_feedback": True,
            },
            ProtocolConstants.TRANSCRIPTION: {
                "function": get_levenshtein_feedback,
                "files": [ProtocolConstants.TRANSCRIPTION],
                "columns": [0],
                "detection_req": ProtocolConstants.SKIP,
                "budgeted_feedback": True,
            },
            ProtocolConstants.SCORE: {
                "function": get_classificaton_score_feedback,
                "files": [ProtocolConstants.CLASSIFICATION],
                "columns": [4],
                "detection_req": ProtocolConstants.SKIP,
                "budgeted_feedback": False,
            },
        },
        "activity_recognition": {
            ProtocolConstants.CLASSIFICATION: {
                "function": get_classification_var_feedback,
                "files": [ProtocolConstants.CLASSIFICATION],
                "columns": [5, 10],
                "detection_req": ProtocolConstants.SKIP,
                "budgeted_feedback": True,
            },
            ProtocolConstants.SCORE: {
                "function": get_classificaton_score_feedback,
                "files": [ProtocolConstants.CLASSIFICATION],
                "columns": [2],
                "detection_req": ProtocolConstants.SKIP,
                "budgeted_feedback": False,
            },
            ProtocolConstants.DETECTION: {
                "function": get_detection_feedback,
                "files": [ProtocolConstants.DETECTION],
                "columns": [0],
                "detection_req": ProtocolConstants.SKIP,
                "budgeted_feedback": True,
                "required_hints": [],
                "alternate_budget": "max_detection_feedback_ids",
            },
        },
    }

    def get_feedback(
        self, feedback_ids: List[str], feedback_type: str, session_id: str, test_id: str
    ) -> BytesIO:
        """
        Get feedback of the specified type.

        Args:
            feedback_ids: List of media ids for which feedback is required
            feedback_type: Protocols constants with the values: label, detection, characterization
            session_id: The id provided by a server denoting a session
            test_id: The id of the test currently being evaluated

        Returns:
            An instance of BytesIO with feedback
        """
        metadata = self.get_test_metadata(session_id, test_id, False)
        structure = get_session_info(self.results_folder, session_id)
        test_structure = get_session_test_info(self.results_folder, session_id, test_id)
        domain = structure["created"]["domain"]
        if domain not in self.feedback_request_mapping:
            raise ProtocolError(
                "BadDomain",
                f"The set domain does not match a domain type. Please check the metadata file for {test_id}",
                "".join(traceback.format_stack()),
            )

        # Ensure feedback type works with session domain
        # and if so, grab the proper files

        try:
            feedback_definition = self.feedback_request_mapping[domain][feedback_type]
            file_types = feedback_definition["files"]
        except Exception:
            raise ProtocolError(
                "InvalidFeedbackType",
                f"Invalid feedback type requested for the test id {test_id} with domain {domain}",
                "".join(traceback.format_stack()),
            )

        is_given_detection_mode = "red_light" in structure["created"].get("hints", [])
        budgeted_feedback = feedback_definition["budgeted_feedback"] and not (
            feedback_type == ProtocolConstants.DETECTION and is_given_detection_mode
        )

        if (
            "alternate_budget" in feedback_definition
            and feedback_definition["alternate_budget"] in metadata
        ):
            feedback_budget = int(metadata[feedback_definition["alternate_budget"]])
        else:
            feedback_budget = int(metadata.get("feedback_max_ids", 0))

        try:
            # Gets the amount of ids already requested for this type of feedback this round and
            # determines whether the limit has already been reached
            feedback_round_id = str(
                max([int(r) for r in test_structure["post_results"]["rounds"].keys()])
            )

            feedback_count = test_structure["get_feedback"]["rounds"][
                feedback_round_id
            ].get(feedback_type, 0)
            if feedback_count >= feedback_budget:
                raise ProtocolError(
                    "FeedbackBudgetExceeded",
                    f"Feedback of type {feedback_type} has already been requested on the maximum number of ids",
                )
        except KeyError:
            feedback_round_id = str(0)
            feedback_count = 0

        ground_truth_file = os.path.join(
            self.folder, metadata["protocol"], domain, f"{test_id}_single_df.csv"
        )

        if not os.path.exists(ground_truth_file):
            raise ServerError(
                "test_id_invalid",
                f"Could not find ground truth file for test Id {test_id}",
                "".join(traceback.format_stack()),
            )

        results_files = []
        for t in file_types:
            results_files.append(
                os.path.join(
                    self.results_folder,
                    metadata["protocol"],
                    domain,
                    f"{str(session_id)}.{str(test_id)}_{t}.csv",
                )
            )

        if len(results_files) < len(file_types):
            raise ServerError(
                "test_id_invalid",
                f"Could not find posted result file(s) for test Id {test_id} with feedback type {feedback_type}",
                "".join(traceback.format_stack()),
            )

        # Ensure any required hint(s) are present in the session info structure
        req_hints = feedback_definition.get("required_hints", [])
        if len(req_hints) > 0:
            for hint in req_hints:
                if hint not in structure["created"].get("hints", []):
                    logging.warning(
                        "Inform TA2 team that they are requesting feedback prior to the threshold indication"
                    )
                    return BytesIO()

        detection_requirement = feedback_definition.get(
            "detection_req", ProtocolConstants.IGNORE
        )

        # If novelty detection is required, ensure detection has been posted
        # for the requested round and novelty claimed for the test
        if detection_requirement != ProtocolConstants.IGNORE:
            test_results_structure = test_structure["post_results"]
            if "detection file path" not in test_results_structure:
                raise ProtocolError(
                    "DetectionPostRequired",
                    """A detection file is required to be posted before feedback can be requested on a round.
                       Please submit Detection results before requesting feedback""",
                )
            else:
                try:
                    with open(
                        test_results_structure["detection file path"], "r"
                    ) as d_file:
                        d_reader = csv.reader(d_file, delimiter=",")
                        detection_lines = list(d_reader)
                    predictions = [float(x[1]) for x in detection_lines]
                    # if given detection and past the detection point
                    is_given = is_given_detection_mode and metadata.get(
                        "red_light"
                    ) in [x[0] for x in detection_lines]
                    if (
                        max(predictions) <= structure["created"]["detection_threshold"]
                        and not is_given
                    ):
                        if (
                            detection_requirement
                            == ProtocolConstants.NOTIFY_AND_CONTINUE
                        ):
                            logging.error(
                                "Inform TA2 team that they are requesting feedback prior to the threshold indication"
                            )
                        elif detection_requirement == ProtocolConstants.SKIP:
                            logging.warning(
                                "Inform TA2 team that they are requesting feedback prior to the threshold indication"
                            )
                            return BytesIO()
                        else:
                            raise ProtocolError(
                                "NoveltyDetectionRequired",
                                "In order to request feedback, novelty must be declared for the test",
                            )
                except ProtocolError as e:
                    raise e
                except Exception:
                    raise ServerError(
                        "CantReadFile",
                        f"""Couldnt open the logged detection file at
                            {test_results_structure['detection file path']}.
                            Please check if the file exists and that
                            {session_id}.json has the proper file location
                            for test id {test_id}""",
                        traceback.format_exc(),
                    )

        # Add columns to metadata for use in feedback
        metadata["columns"] = feedback_definition.get("columns", [])

        # Get feedback from specified test
        try:
            if "psuedo" in feedback_type:
                feedback = psuedo_label_feedback(
                    ground_truth_file,
                    feedback_ids,
                    feedback_definition["files"][0],
                    metadata,
                    self.results_folder,
                    session_id,
                )
            else:
                feedback = feedback_definition["function"](
                    ground_truth_file, results_files, feedback_ids, metadata
                )
        except KeyError:
            raise ProtocolError(
                "feedback_type_invalid",
                f"""Feedback type {feedback_type} is not valid. Make sure the
                    provider's feedback_algorithms variable is properly set up""",
                traceback.format_exc(),
            )

        number_of_ids_to_return = len(feedback)

        # if budgeted, decrement use and check if too many has been requested
        if budgeted_feedback:
            left_over_ids = feedback_budget - feedback_count
            number_of_ids_to_return = min(number_of_ids_to_return, left_over_ids)
        feedback_count += number_of_ids_to_return

        log_session(
            self.results_folder,
            session_id=session_id,
            activity="get_feedback",
            test_id=test_id,
            round_id=int(feedback_round_id),
            content={feedback_type: feedback_count, "feedback_budget": feedback_budget},
        )

        feedback_csv = BytesIO()
        for key in feedback.keys():
            if type(feedback[key]) is not list:
                feedback_csv.write(f"{key},{feedback[key]}\n".encode("utf-8"))
            else:
                feedback_csv.write(
                    f"{key},{','.join(str(x) for x in feedback[key])}\n".encode("utf-8")
                )
            number_of_ids_to_return -= 1
            # once maximium requested number is hit, quit
            if number_of_ids_to_return == 0:
                break

        feedback_csv.seek(0)

        return feedback_csv

    def post_results(
        self,
        session_id: str,
        test_id: str,
        round_id: int,
        result_files: Dict[str, str],
    ) -> None:
        """
        Post results.

        Args:
            result_files: A dictionary of results with protocol constant as key and file path as value
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            None
        """
        # Modify session file with posted results
        structure = get_session_info(self.results_folder, session_id)
        test_structure = get_session_test_info(self.results_folder, session_id, test_id)
        if "detection" in result_files.keys():
            try:
                if (
                    "detection"
                    in test_structure["post_results"]["rounds"][str(round_id)]["types"]
                ):
                    raise ProtocolError(
                        "DetectionRepost",
                        """Cannot re post detection for a given round. If you
                           attempted to submit any other results, please
                           resubmit without detection.""",
                    )
            except KeyError:
                pass

        protocol = structure["created"]["protocol"]
        domain = structure["created"]["domain"]
        os.makedirs(os.path.join(self.results_folder, protocol, domain), exist_ok=True)
        log_content = {}
        for r_type in result_files.keys():
            filename = f"{str(session_id)}.{str(test_id)}_{r_type}.csv"
            path = os.path.join(self.results_folder, protocol, domain, filename)
            log_content[f"{r_type} file path"] = path
            with open(path, "a+") as result_file:
                result_file.write(result_files[r_type])

        # Log call
        log_content["last round"] = str(round_id)
        updated_test_structure = log_session(
            self.results_folder,
            activity="post_results",
            session_id=session_id,
            test_id=test_id,
            round_id=round_id,
            content=log_content,
            content_loc="activity",
            return_structure=True,
        )
        if updated_test_structure:
            prev_types = updated_test_structure["post_results"]["rounds"][
                str(round_id)
            ].get("types", [])
            new_types = prev_types + list(result_files.keys())
            updated_test_structure["post_results"]["rounds"][str(round_id)][
                "types"
            ] = new_types
            write_session_log_file(
                updated_test_structure,
                os.path.join(
                    self.results_folder, f"{str(session_id)}.{str(test_id)}.json"
                ),
            )

    def complete_test(self, session_id: str, test_id: str) -> None:
        """
        Mark test as completed in session logs.

        Args:
            session_id: The id of the session currently being evaluated
            test_id: The id of the test currently being evaluated

        Returns:
            None
        """
        log_session(
            self.results_folder,
            session_id=session_id,
            test_id=test_id,
            activity="completion",
        )

    def terminate_session(self, session_id: str) -> None:
        """
        Terminate the session.

        Args:
            session_id: The id provided by a server denoting a session

        Returns: None
        """
        # Modify session file to indicate session has been terminated
        log_session(self.results_folder, session_id=session_id, activity="termination")

    def latest_session_info(self, session_id: str) -> Dict:
        """
        Get tests finished from the most recent session.

        Args:
            session_id: The id of the session currently being evaluated

        Returns:
            List of tests that have been marked as completed in the session
        """
        structure = get_session_info(self.results_folder, session_id)
        latest = {}
        latest["finished_tests"] = structure["tests"]["completed_tests"]
        return latest



"""Program Metric Functions."""

import warnings
import numpy as np
from sail_on_client.evaluate.utils import (
    check_novel_validity,
    check_class_validity,
    get_first_detect_novelty,
    top1_accuracy,
    top3_accuracy,
    get_rolling_stats,
    topk_accuracy,
)
from typing import Dict


DETECT_THRESH_ = [0.175, 0.225, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]


def _get_threshold(p_novel: np.ndarray) -> np.ndarray:
    return np.min(p_novel) + ((np.max(p_novel) - np.min(p_novel)) / 2.0)


def m_num(p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
    """
    Program Metric: Number of samples needed for detecting novelty.

    The method computes number of GT novel samples needed to predict the first true positive.

    Args:
        p_novel: NX1 vector with each element corresponding to probability of novelty
        gt_novel: NX1 vector with each element 0 (not novel) or 1 (novel)

    Returns:
        single scalar for number of GT novel samples
    """
    res = {}
    check_novel_validity(p_novel, gt_novel)
    if np.sum(gt_novel) < 1:
        for thresh in DETECT_THRESH_:
            res[f"{thresh}"] = -1
    else:
        for thresh in DETECT_THRESH_:
            res[f"{thresh}"] = ((p_novel[gt_novel == 1] >= thresh).argmax(axis=0)) + 1
    return res


def m_num_stats(p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
    """
    Program Metric: Number of samples needed for detecting novelty.

    The method computes number of GT novel samples needed to predict the first true positive.

    Args:
        p_novel: NX1 vector with each element corresponding to probability of novelty
        gt_novel: NX1 vector with each element 0 (not novel) or 1 (novel)

    Returns:
        single scalar for number of GT novel samples
    """
    check_novel_validity(p_novel, gt_novel)
    res = {}
    if np.sum(gt_novel) < 1:
        # only print warning first time.
        first_novel_indx = len(gt_novel) + 1
    else:
        first_novel_indx = np.where(gt_novel == 1)[0][0] + 1
    res["GT_indx"] = first_novel_indx

    for thresh in DETECT_THRESH_:
        res[f"P_indx_{thresh}"] = get_first_detect_novelty(p_novel, thresh)
    return res


def m_ndp(p_novel: np.ndarray, gt_novel: np.ndarray, mode: str = "full_test") -> Dict:
    """
    Program Metric: Novelty detection performance.

    The method computes per-sample novelty detection performance

    Args:
        p_novel: NX1 vector with each element corresponding to probability of it being novel
        gt_novel: NX1 vector with each element 0 (not novel) or 1 (novel)
        mode: if 'full_test' computes on all test samples, if 'post_novelty' computes from first GT novel sample

    Returns:
        Dictionary of various metrics: Accuracy, Precision, Recall, F1_score and Confusion matrix
    """
    check_novel_validity(p_novel, gt_novel)
    if mode not in ["full_test", "pre_novelty", "post_novelty"]:
        raise Exception(
            "Mode should be one of ['full_test','pre_novelty', 'post_novelty']"
        )

    def get_all_scores_ndp(thresh: float) -> Dict:
        preds: np.ndarray = p_novel > thresh
        gt_novel_ = gt_novel
        if mode == "post_novelty":
            if any(gt_novel != 0):
                post_novel_idx = (gt_novel != 0).argmax(axis=0)
                preds = preds[post_novel_idx:]
                gt_novel_ = gt_novel[post_novel_idx:]
            else:
                return {
                    f"accuracy_{thresh}": -1,
                    f"precision_{thresh}": -1,
                    f"recall_{thresh}": -1,
                    f"f1_score_{thresh}": -1,
                    f"TP_{thresh}": -1,
                    f"FP_{thresh}": -1,
                    f"TN_{thresh}": -1,
                    f"FN_{thresh}": -1,
                }
        elif mode == "pre_novelty":
            if any(gt_novel != 0) and not all(gt_novel != 0):
                post_novel_idx = (gt_novel != 0).argmax(axis=0)
                preds = preds[:post_novel_idx]
                gt_novel_ = gt_novel[:post_novel_idx]

        tp = np.sum(np.logical_and(preds == 1, gt_novel_ == 1))
        fp = np.sum(np.logical_and(preds == 1, gt_novel_ == 0))
        tn = np.sum(np.logical_and(preds == 0, gt_novel_ == 0))
        fn = np.sum(np.logical_and(preds == 0, gt_novel_ == 1))

        acc = (tp + tn) / (tp + tn + fp + fn)
        if tp + fp == 0.0:
            precision = 0.0
        else:
            precision = tp / (tp + fp)
        if tp + fn == 0.0:
            recall = 0.0
        else:
            recall = tp / (tp + fn)
        if precision == 0.0 and recall == 0.0:
            f1_score = 0.0
        else:
            f1_score = 2 * precision * recall / (precision + recall)

        return {
            f"accuracy_{thresh}": round(acc, 5),
            f"precision_{thresh}": round(precision, 5),
            f"recall_{thresh}": round(recall, 5),
            f"f1_score_{thresh}": round(f1_score, 5),
            f"TP_{thresh}": tp,
            f"FP_{thresh}": fp,
            f"TN_{thresh}": tn,
            f"FN_{thresh}": fn,
        }

    res = {}
    for thresh in DETECT_THRESH_:
        res.update(get_all_scores_ndp(thresh))
    return res


def m_acc(
    gt_novel: np.ndarray,
    p_class: np.ndarray,
    gt_class: np.ndarray,
    round_size: int,
    asymptotic_start_round: int,
) -> Dict:
    """
    Compute top1 and top3 accuracy.

    Args:
        p_novel: NX1 vector with each element corresponding to probability of novelty
        p_class : Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
        gt_class : Nx1 vector with ground-truth class for each sample
        round_size: Number of samples in a single round of the test
        asymptotic_start_round: Round id where metric computation starts

    Returns:
        Dictionary with results
    """
    # full test
    batch_size = round_size
    results = {}
    try:
        results["full_top1"] = top1_accuracy(p_class, gt_class, txt="full_top1")
        results["full_top3"] = top3_accuracy(p_class, gt_class, txt="full_top3")

        # red button push:
        if np.sum(gt_novel) < 1:
            # only print warning first time.
            first_novel_indx = len(gt_novel) + 1
        else:
            first_novel_indx = np.where(gt_novel == 1)[0][0] + 1

        if first_novel_indx == len(gt_novel) + 1:
            results["pre_top1"] = top1_accuracy(p_class, gt_class, txt="pre_top1")
            results["pre_top3"] = top3_accuracy(p_class, gt_class, txt="pre_top3")
            results["post_top1"] = -1
            results["post_top3"] = -1
            results["post_mean_top1"] = -1
            results["post_mean_top3"] = -1
            results["post_std_top1"] = -1
            results["post_std_top3"] = -1
        else:
            # pre_novelty
            if first_novel_indx == 0:
                results["pre_top1"] = -1
                results["pre_top3"] = -1
                results["pre_mean_top1"] = -1
                results["pre_mean_top3"] = -1
                results["pre_std_top1"] = -1
                results["pre_std_top3"] = -1
            else:
                p_class_pre = p_class[:first_novel_indx]
                gt_class_pre = gt_class[:first_novel_indx]
                results["pre_top1"] = top1_accuracy(
                    p_class_pre, gt_class_pre, txt="pre_top1"
                )
                results["pre_top3"] = top3_accuracy(
                    p_class_pre, gt_class_pre, txt="pre_top3"
                )
                [results["pre_mean_top1"], results["pre_std_top1"]] = get_rolling_stats(
                    p_class_pre, gt_class_pre, k=1, window_size=batch_size
                )
                [results["pre_mean_top3"], results["pre_std_top3"]] = get_rolling_stats(
                    p_class_pre, gt_class_pre, k=3, window_size=batch_size
                )

            # post_novelty
            p_class_post = p_class[first_novel_indx:]
            gt_class_post = gt_class[first_novel_indx:]
            results["post_top1"] = top1_accuracy(
                p_class_post, gt_class_post, txt="post_top1"
            )
            results["post_top3"] = top3_accuracy(
                p_class_post, gt_class_post, txt="post_top3"
            )
            [results["post_mean_top1"], results["post_std_top1"]] = get_rolling_stats(
                p_class_post, gt_class_post, k=1, window_size=batch_size
            )
            [results["post_mean_top3"], results["post_std_top3"]] = get_rolling_stats(
                p_class_post, gt_class_post, k=3, window_size=batch_size
            )

        # asymptotic performance
        for last_i in np.arange(
            int(asymptotic_start_round) * batch_size, gt_novel.shape[0], round_size
        ):
            if len(gt_novel) > last_i:
                p_class_asym = p_class[-last_i:]
                gt_class_asym = gt_class[-last_i:]
                results[f"asymptotic_{last_i}_top1"] = top1_accuracy(
                    p_class_asym, gt_class_asym, txt=f"asymptotic_{last_i}_top1"
                )
                results[f"asymptotic_{last_i}_top3"] = top3_accuracy(
                    p_class_asym, gt_class_asym, txt=f"asymptotic_{last_i}_top3"
                )
                [
                    results[f"asymptotic_{last_i}_mean_top1"],
                    results[f"asymptotic_{last_i}_std_top1"],
                ] = get_rolling_stats(
                    p_class_asym, gt_class_asym, k=1, window_size=batch_size
                )
                [
                    results[f"asymptotic_{last_i}_mean_top3"],
                    results[f"asymptotic_{last_i}_std_top3"],
                ] = get_rolling_stats(
                    p_class_asym, gt_class_asym, k=3, window_size=batch_size
                )
            else:
                results[f"asymptotic_{last_i}_top1"] = -1
                results[f"asymptotic_{last_i}_top3"] = -1
                results[f"asymptotic_{last_i}_mean_top1"] = -1
                results[f"asymptotic_{last_i}_mean_top3"] = -1
                results[f"asymptotic_{last_i}_std_top1"] = -1
                results[f"asymptotic_{last_i}_std_top3"] = -1
    except Exception:
        import code
        import traceback as tb

        tb.print_stack()
        namespace = globals().copy()
        namespace.update(locals())
        code.interact(local=namespace)

    return results


def m_ndp_pre(p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
    """
    Additional Metric: Novelty detection performance before novelty is introduced.

    Args:
        p_novel: NX1 vector with each element corresponding to probability of novelty
        gt_novel: NX1 vector with each element 0 (not novel) or 1 (novel)

    Returns:
        Dictionary of following metrics values: Accuracy, Precision, Recall, F1_score and Confusion matrix
    """
    return m_ndp(p_novel, gt_novel, mode="pre_novelty")


def m_ndp_post(p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
    """
    Additional Metric: Novelty detection performance after novelty is introduced.

    Args:
        p_novel: NX1 vector with each element corresponding to probability of novelty
        gt_novel: NX1 vector with each element 0 (not novel) or 1 (novel)

    Returns:
        Dictionary of following metrics values: Accuracy, Precision, Recall, F1_score and Confusion matrix
    """
    return m_ndp(p_novel, gt_novel, mode="post_novelty")


def m_ndp_failed_reaction(
    p_novel: np.ndarray,
    gt_novel: np.ndarray,
    p_class: np.ndarray,
    gt_class: np.ndarray,
    mode: str = "full_test",
) -> Dict:
    """
    Additional Metric: Novelty detection when reaction fails.

    The method computes novelty detection performance for only on samples with incorrect k-class predictions

    Args:
        p_novel: NX1 vector with each element corresponding to probability of novelty
        gt_novel: NX1 vector with each element 0 (not novel) or 1 (novel)
        p_class : Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
        gt_class : Nx1 vector with ground-truth class for each sample
        mode: if 'full_test' computes on all test samples, if 'post_novelty' computes from the first GT novel sample
        k: 'k' used in top-K accuracy

    Returns:
        Dictionary of various metrics: Accuracy, Precision, Recall, F1_score and Confusion matrix
    """
    check_class_validity(p_class, gt_class)
    results = {}
    for k in [1, 3]:
        p_class_k = np.argsort(-p_class, axis=1)[:, :k]
        gt_class_k = gt_class[:, np.newaxis]
        check_zero = p_class_k - gt_class_k
        incorrect_mask = ~np.any(check_zero == 0, axis=1)
        if np.sum(incorrect_mask) == 0:
            warnings.warn(
                "WARNING! No incorrect predictions found. Returning empty dictionary"
            )
            for metric in {
                "accuracy",
                "precision",
                "recall",
                "f1_score",
                "FN",
                "TP",
                "FP",
                "TN",
            }:
                results[f"top{k}_{metric}"] = -1
            continue
        p_novel_k = p_novel[incorrect_mask]
        gt_novel_k = gt_novel[incorrect_mask]
        res = m_ndp(p_novel_k, gt_novel_k, mode)
        for metric in res:
            results[f"top{k}_{metric}"] = res[metric]
    return results


def m_accuracy_on_novel(
    p_class: np.ndarray, gt_class: np.ndarray, gt_novel: np.ndarray
) -> Dict:
    """
    Additional Metric: Novelty robustness.

    The method computes top-K accuracy for only the novel samples

    Args:
        p_class : Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
        gt_class : Nx1 vector with ground-truth class for each sample
        gt_novel : Nx1 binary vector corresponding to the ground truth novel{1}/seen{0} labels
        k : K value to compute accuracy at

    Returns:
        Accuracy at rank-k
    """
    check_class_validity(p_class, gt_class)
    if np.sum(gt_novel) < 1:
        return {"top3_acc_novel_only": -1, "top1_acc_novel_only": -1}
    p_class = p_class[gt_novel == 1]
    gt_class = gt_class[gt_novel == 1]
    return {
        "top3_acc_novel_only": topk_accuracy(p_class, gt_class, k=3),
        "top1_acc_novel_only": topk_accuracy(p_class, gt_class, k=1),
    }



"""CONDDA protocol."""

from sail_on_client.agent.condda_agent import CONDDAAgent
from sail_on_client.harness.test_and_evaluation_harness import (
    TestAndEvaluationHarnessType,
)
from sail_on_client.protocol.visual_protocol import VisualProtocol
from sail_on_client.protocol.condda_dataclasses import AlgorithmAttributes
from sail_on_client.protocol.condda_test import CONDDATest

import logging

from typing import Dict, List

log = logging.getLogger(__name__)


class Condda(VisualProtocol):
    """CONDDA protocol."""

    def __init__(
        self,
        algorithms: Dict[str, CONDDAAgent],
        dataset_root: str,
        domain: str,
        harness: TestAndEvaluationHarnessType,
        save_dir: str,
        seed: str,
        test_ids: List[str],
        feature_extraction_only: bool = False,
        hints: List = None,
        is_eval_enabled: bool = False,
        is_eval_roundwise_enabled: bool = False,
        resume_session: bool = False,
        resume_session_ids: Dict = None,
        save_attributes: bool = False,
        saved_attributes: Dict = None,
        save_elementwise: bool = False,
        save_features: bool = False,
        feature_dir: str = "",
        skip_stages: List = None,
        use_consolidated_features: bool = False,
        use_saved_attributes: bool = False,
        use_saved_features: bool = False,
    ) -> None:
        """
        Initialize CONDDA protocol object.

        Args:
            algorithms: Dictionary of algorithms that are used run based on the protocol
            dataset_root: Root directory of the dataset
            domain: Domain of the problem
            save_dir: Directory where results are saved
            seed: Seed for the experiments
            test_ids: List of tests
            feature_extraction_only: Flag to only run feature extraction
            hints: List of hint provided in the session
            harness: A harness for test and evaluation
            is_eval_enabled: Flag to check if evaluation is enabled in session
            is_eval_roundwise_enabled: Flag to check if evaluation is enabled for rounds
            resume_session: Flag to resume session
            resume_session_ids: Dictionary for resuming sessions
            save_attributes: Flag to save attributes
            saved_attributes: Dictionary for attributes
            save_elementwise: Flag to save features elementwise
            save_features: Flag to save  features
            feature_dir: Directory to save features
            skip_stages: List of stages that are skipped
            use_consolidated_features: Flag to use consolidated features
            use_saved_attributes: Flag to use saved attributes
            use_saved_features: Flag to use saved features

        Returns:
            None
        """
        super().__init__(algorithms, harness)
        self.dataset_root = dataset_root
        self.domain = domain
        self.feature_extraction_only = feature_extraction_only
        self.feature_dir = feature_dir
        if hints is None:
            self.hints = []
        else:
            self.hints = hints
        self.is_eval_enabled = is_eval_enabled
        self.is_eval_roundwise_enabled = is_eval_roundwise_enabled
        self.resume_session = resume_session
        if resume_session_ids is None:
            self.resume_session_ids = {}
        else:
            self.resume_session_ids = resume_session_ids
        self.save_attributes = save_attributes
        if saved_attributes is None:
            self.saved_attributes = {}
        else:
            self.saved_attributes = saved_attributes
        self.save_dir = save_dir
        self.save_elementwise = save_elementwise
        self.save_features = save_features
        if skip_stages is None:
            self.skip_stages = []
        else:
            self.skip_stages = skip_stages
        self.seed = seed
        self.test_ids = test_ids
        self.use_consolidated_features = use_consolidated_features
        self.use_saved_attributes = use_saved_attributes
        self.use_saved_features = use_saved_features

    def create_algorithm_attributes(
        self, algorithm_name: str, algorithm_param: Dict, test_ids: List[str]
    ) -> AlgorithmAttributes:
        """
        Create an instance of algorithm attributes.

        Args:
            algorithm_name: Name of the algorithm
            algorithm_param: Parameters for the algorithm
            baseline_algorithm_name: Name of the baseline algorithm
            has_baseline: Flag to check if a baseline is present in the config
            has_reaction_baseline: Flag to check if a reaction baseline is present in the config
            test_ids: List of test

        Returns:
            An instance of AlgorithmAttributes
        """
        algorithm_instance = self.algorithms[algorithm_name]
        session_id = self.resume_session_ids.get(algorithm_name, "")
        return AlgorithmAttributes(
            algorithm_name,
            algorithm_param.get("detection_threshold", 0.5),
            algorithm_instance,
            algorithm_param.get("package_name", None),
            algorithm_param,
            session_id,
            test_ids,
        )

    def create_algorithm_session(
        self,
        algorithm_attributes: AlgorithmAttributes,
        domain: str,
        hints: List[str],
        has_a_session: bool,
        protocol_name: str,
    ) -> AlgorithmAttributes:
        """
        Create/resume session for an algorithm.

        Args:
            algorithm_attributes: An instance of AlgorithmAttributes
            domain: Domain for the algorithm
            hints: List of hints used in the session
            has_a_session: Already has a session and we want to resume it
            protocol_name: Name of the algorithm

        Returns:
            An AlgorithmAttributes object with updated session id or test id
        """
        test_ids = algorithm_attributes.test_ids
        named_version = algorithm_attributes.named_version()
        detection_threshold = algorithm_attributes.detection_threshold

        if has_a_session:
            session_id = algorithm_attributes.session_id
            finished_test = self.harness.resume_session(session_id)
            algorithm_attributes.remove_completed_tests(finished_test)
            log.info(f"Resumed session {session_id} for {algorithm_attributes.name}")
        else:
            session_id = self.harness.session_request(
                test_ids,
                protocol_name,
                domain,
                named_version,
                hints,
                detection_threshold,
            )
            algorithm_attributes.session_id = session_id
            log.info(f"Created session {session_id} for {algorithm_attributes.name}")
        return algorithm_attributes

    def update_skip_stages(
        self, skip_stages: List[str], save_features: bool, feature_extraction_only: bool
    ) -> List[str]:
        """
        Update skip stages based on the boolean values in config.

        Args:
            skip_stages: List of skip stages specified in the config
            is_eval_enabled: Flag to enable evaluation
            is_eval_roundwise_enabled: Flag to enable evaluation in every round
            use_feedback: Flag to enable using feedback
            save_features: Flag to enable saving features
            feature_extraction_only: Flag to only run feature extraction

        Returns:
            Update list of skip stages
        """
        if not save_features:
            skip_stages.append("SaveFeatures")
        if feature_extraction_only:
            skip_stages.append("WorldDetection")
            skip_stages.append("NoveltyCharacterization")
        return skip_stages

    def run_protocol(self, config: Dict) -> None:
        """
        Run protocol.

        Args:
            config: Parameters provided in the config

        Returns:
            None
        """
        log.info("Starting CONDDA")
        self.skip_stages = self.update_skip_stages(
            self.skip_stages, self.save_features, self.feature_extraction_only
        )
        algorithms_attributes = []
        # Populate most of the attributes for the algorithm
        for algorithm_name in self.algorithms.keys():
            algorithm_param = self.algorithms[algorithm_name].get_config()
            algorithm_attributes = self.create_algorithm_attributes(
                algorithm_name, algorithm_param, self.test_ids
            )
            log.info(f"Consolidating attributes for {algorithm_name}")
            algorithms_attributes.append(algorithm_attributes)

        # Create sessions an instances of all the algorithms and populate
        # session_id for algorithm attributes
        for idx, algorithm_attributes in enumerate(algorithms_attributes):
            algorithms_attributes[idx] = self.create_algorithm_session(
                algorithm_attributes,
                self.domain,
                self.hints,
                self.resume_session,
                "CONDDA",
            )

        # Run tests for all the algorithms
        for algorithm_attributes in algorithms_attributes:
            algorithm_name = algorithm_attributes.name
            session_id = algorithm_attributes.session_id
            test_ids = algorithm_attributes.test_ids
            log.info(f"Starting session: {session_id} for algorithm: {algorithm_name}")
            skip_stages = self.skip_stages.copy()
            condda_test = CONDDATest(
                algorithm_attributes,
                self.dataset_root,
                self.domain,
                "",
                self.harness,
                self.save_dir,
                session_id,
                skip_stages,
                self.use_consolidated_features,
                self.use_saved_features,
            )
            for test_id in test_ids:
                log.info(f"Start test: {test_id}")
                condda_test(test_id)
                log.info(f"Test complete: {test_id}")

        # Terminate algorithms
        for algorithm_attributes in algorithms_attributes:
            algorithm_name = algorithm_attributes.name
            session_id = algorithm_attributes.session_id
            self.harness.terminate_session(session_id)
            log.info(f"Session ended for {algorithm_name}: {session_id}")



"""Exception objects representing HTTP error conditions."""

from typing import Any, List, Tuple, Dict


def _get_all_subclasses(cls: Any) -> List[Any]:
    from itertools import chain

    return list(
        chain.from_iterable(
            [
                list(chain.from_iterable([[x], _get_all_subclasses(x)]))
                for x in cls.__subclasses__()
            ]
        )
    )


class ApiError(Exception):
    """Base class for all server side error conditions."""

    reason = "Unknown"
    msg = ""
    stack_trace = "stack trace unavailable"
    error_code = 500

    def __init__(
        self, reason: str, msg: str, stack_trace: str = "stack trace unavailable"
    ):
        """
        Initialize the api error object.

        Args:
            reason: Cause for the error
            msg: Additional message associated with the error
            stack_trace: Stack trace associated with the error

        Returns:
            None
        """
        self.reason = reason
        self.msg = msg
        self.stack_trace = stack_trace

    @staticmethod
    def error_classes() -> List[Any]:
        """
        Return all error classes in the system.

        Returns:
            List of subclasses of server error
        """
        return _get_all_subclasses(ApiError)

    def flask_response(self) -> Tuple[Dict[str, str], int]:
        """
        Convert the object to a Flask response.

        Returns:
            Tuple of response obtained from server and error code
        """
        response = {
            "reason": self.reason,
            "message": self.msg,
            "stack_trace": self.stack_trace,
        }

        return response, self.error_code


class ServerError(ApiError):
    """500-series error indicating general problem on the serverside."""

    error_code = 500

    def __init__(
        self, reason: str, msg: str, stack_trace: str = "stack trace unavailable"
    ):
        """
        Initialize the server error object.

        Args:
            reason: Cause for the error
            msg: Additional message associated with the error
            stack_trace: Stack trace associated with the error

        Returns:
            None
        """
        super().__init__(reason, msg, stack_trace)


class RoundError(ServerError):
    """Error indicating problem with rounds."""

    error_code = 204

    def __init__(
        self, reason: str, msg: str, stack_trace: str = "stack trace unavailable"
    ):
        """
        Initialize the round error object.

        Args:
            reason: Cause for the error
            msg: Additional message associated with the error
            stack_trace: Stack trace associated with the error

        Returns:
            None
        """
        super().__init__(reason, msg, stack_trace)


class ProtocolError(ApiError):
    """400-series error indicating caller caused a problem engaging a protocol."""

    error_code = 400

    def __init__(
        self, reason: str, msg: str, stack_trace: str = "stack trace unavailable"
    ):
        """
        Initialize the Protocol error object.

        Args:
            reason: Cause for the error
            msg: Additional message associated with the error
            stack_trace: Stack trace associated with the error

        Returns:
            None
        """
        super().__init__(reason, msg, stack_trace)



"""Reaction agent that use precomputed values in the OND protocol."""

from sail_on_client.agent.ond_reaction_agent import ONDReactionAgent
from sail_on_client.agent.pre_computed_detector import PreComputedONDAgent

import logging
from typing import Dict, Any, Tuple

log = logging.getLogger(__name__)


class PreComputedONDReactionAgent(ONDReactionAgent):
    """Detector for submitting precomputed results for computing reaction performance."""

    def __init__(
        self,
        algorithm_name: str,
        cache_dir: str,
        has_roundwise_file: bool,
        round_size: int,
    ) -> None:
        """
        Construct agent with precomputed results.

        Args:
            algorithm_name: Name of the algorithm
            cache_dir: Path to cache directory
            has_roundwise_file: Flag to determine if the cache has files for rounds
            round_size: Size of a round
        """
        self.detector = PreComputedONDAgent(
            algorithm_name, cache_dir, has_roundwise_file, round_size
        )

    def get_config(self) -> Dict:
        """Return a default configuration dictionary."""
        return {
            "algorithm_name": self.detector.algorithm_name,
            "cache_dir": self.detector.cache_dir,
            "has_roundwise_file": self.detector.has_roundwise_file,
            "round_size": self.detector.round_size,
        }

    def execute(self, toolset: Dict, step_descriptor: str) -> Any:
        """
        Execute method used by the protocol to run different steps associated with the algorithm.

        Args:
            toolset (dict): Dictionary containing parameters for different steps
            step_descriptor (str): Name of the step
        """
        log.info(f"Executing {step_descriptor}")
        return self.detector.step_dict[step_descriptor](toolset)

    def initialize(self, toolset: Dict) -> None:
        """
        Algorithm Initialization.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            None
        """
        self.detector.initialize(toolset)

    def feature_extraction(
        self, toolset: Dict
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Feature extraction step for the algorithm.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            Tuple of dictionary
        """
        return self.detector.feature_extraction(toolset)

    def novelty_classification(self, toolset: Dict) -> str:
        """
        Classify data provided in known classes and unknown class.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            path to csv file containing the results for novelty classification step
        """
        return self.detector.novelty_classification(toolset)



"""Round for OND."""

import logging
from typing import List, Any, Dict, Union

from sail_on_client.protocol.ond_dataclasses import (
    NoveltyClassificationParams,
    NoveltyAdaptationParams,
)
from sail_on_client.protocol.visual_dataclasses import (
    FeatureExtractionParams,
    WorldChangeDetectionParams,
)
from sail_on_client.protocol.visual_round import VisualRound
from sail_on_client.harness.test_and_evaluation_harness import (
    TestAndEvaluationHarnessType,
)
from sail_on_client.utils.utils import safe_remove
from sail_on_client.utils.decorators import skip_stage


log = logging.getLogger(__name__)


class ONDRound(VisualRound):
    """Class Representing a round in OND."""

    def __init__(
        self,
        algorithm: Any,
        data_root: str,
        features_dict: Dict,
        harness: TestAndEvaluationHarnessType,
        logit_dict: Dict,
        redlight_instance: str,
        session_id: str,
        skip_stages: List[str],
        test_id: str,
    ) -> None:
        """
        Construct round for OND.

        Args:
            algorithm: An instance of algorithm
            data_root: Root directory of the data
            features_dict: Dictionary with features for the entire dataset
            harness: An instance of the harness used for T&E
            logit_dict: Dictionary with logits for the entire dataset
            redlight_instance: The instance when the world changes
            session_id: Session id associated with the algorithm
            skip_stages: List of stages that are skipped
            test_id: Test id associated with the round

        Returns:
            None
        """
        super().__init__(
            algorithm,
            data_root,
            features_dict,
            harness,
            logit_dict,
            redlight_instance,
            session_id,
            skip_stages,
            test_id,
        )

    @skip_stage("NoveltyClassification")
    def _run_novelty_classification(
        self, nc_params: NoveltyClassificationParams, round_id: int
    ) -> None:
        """
        Private helper function for novelty classification.

        Args:
            nc_params: An instance of dataclass with parameters for novelty classification
            round_id: Identifier for a round

        Returns:
            None
        """
        ncl_result = self.algorithm.execute(
            nc_params.get_toolset(), "NoveltyClassification"
        )
        self.harness.post_results(
            {"classification": ncl_result}, self.test_id, round_id, self.session_id
        )
        safe_remove(ncl_result)

    @skip_stage("EvaluateRoundwise")
    def _evaluate_roundwise(self, round_id: int) -> Dict:
        """
        Compute roundwise accuracy.

        Args:
            round_id: Identifier for a round

        Returns:
            Dictionary with accuracy metrics for round
        """
        return self.harness.evaluate_round_wise(self.test_id, round_id, self.session_id)

    @skip_stage("NoveltyAdaptation")
    def _run_novelty_adaptation(self, na_params: NoveltyAdaptationParams) -> None:
        """
        Private helper function for adaptation.

        Args:
            na_params: An instance of dataclass with parameters for adaptation

        Returns:
            None
        """
        return self.algorithm.execute(na_params.get_toolset(), "NoveltyAdaptation")

    def __call__(self, dataset: str, round_id: int) -> Union[Dict, None]:
        """
        Core logic for running round in OND.

        Args:
            algorithm: An instance of the algorithm
            dataset: Path to a file with the dataset for the round
            round_id: An Identifier for a round

        Returns:
            Score for the round
        """
        # Run feature extraction
        fe_params = FeatureExtractionParams(dataset, self.data_root, round_id)
        instance_ids = ONDRound.get_instance_ids(dataset)
        rfeature_dict, rlogit_dict = self._run_feature_extraction(
            fe_params, instance_ids
        )
        # Run World Change Detection
        wc_params = WorldChangeDetectionParams(
            rfeature_dict, rlogit_dict, round_id, self.redlight_instance
        )
        self._run_world_change_detection(wc_params, round_id)
        # Run Novelty Classification
        nc_params = NoveltyClassificationParams(rfeature_dict, rlogit_dict, round_id)
        self._run_novelty_classification(nc_params, round_id)
        # Compute metrics for the round
        round_score = self._evaluate_roundwise(round_id)

        na_params = NoveltyAdaptationParams(round_id)
        self._run_novelty_adaptation(na_params)

        return round_score



"""Visual Protocol."""

import logging
from typing import Dict, TypeVar, Type, Any

from sail_on_client.harness.test_and_evaluation_harness import (
    TestAndEvaluationHarnessType,
)
from sail_on_client.agent.visual_agent import VisualAgentType
from tinker.protocol import Protocol
from tinker.configuration import process_config


log = logging.getLogger(__name__)
VisualProtocolType = TypeVar("VisualProtocolType", bound="VisualProtocol")


class VisualProtocol(Protocol):
    """Protocol for visual tasks."""

    def __init__(
        self,
        algorithms: Dict[str, VisualAgentType],
        harness: TestAndEvaluationHarnessType,
    ) -> None:
        """
        Construct visual protocol.

        Args:
            algorithms: Dictionary of algorithms that are used run based on the protocol
            harness: A harness for test and evaluation

        Returns:
            None
        """
        super().__init__()
        self.algorithms = algorithms
        self.harness = harness

    @classmethod
    def from_config(
        cls: Type[VisualProtocolType], config_dict: Dict, merge_default: bool = True
    ) -> VisualProtocolType:
        """
        Construct protocol from config.

        Args:
            config_dict: dictionary with parameters
            merge_default: Merge dictionary with default values

        Returns:
            An instance of visual protocol
        """
        config_dict = dict(config_dict)
        algorithm_configs = config_dict.get("algorithms", None)
        if not algorithm_configs:
            raise ValueError("No algorithms provided in the config.")
        for algorithm_name, algorithm_config in algorithm_configs.items():
            config_dict["algorithms"][algorithm_name] = process_config(algorithm_config)
        harness_config = config_dict.get("harness", None)
        if not harness_config:
            raise ValueError("Harness parameters not provided in the config.")
        config_dict["harness"] = process_config(harness_config)
        return super().from_config(config_dict, merge_default=merge_default)

    def get_config(self) -> Dict:
        """
        Get json compliant representation of the protocol.

        Returns:
            Dictionary with json compliant representation of the protocol
        """
        cfg: Dict[str, Any] = {"algorithms": {}}
        for algorithm_name, algorithm in self.algorithms.items():
            cfg["algorithms"][algorithm_name] = {}
            cfg["algorithms"][algorithm_name]["config"] = algorithm.get_config()
            cfg["algorithms"][algorithm_name]["class"] = algorithm.__class__.__name__
        cfg["harness"] = {}
        cfg["harness"]["config"] = self.harness.get_config()
        cfg["harness"]["class"] = self.harness.__class__.__name__
        return cfg



"""Activity Recognition Class for metrics for sail-on."""

from sail_on_client.evaluate.program_metrics import ProgramMetrics
from sail_on_client.evaluate.metrics import m_acc, m_num, m_ndp, m_num_stats
from sail_on_client.evaluate.metrics import m_ndp_failed_reaction
from sail_on_client.evaluate.metrics import m_accuracy_on_novel
from sail_on_client.evaluate.utils import topk_accuracy

import numpy as np
from pandas import DataFrame

from typing import Dict


class ActivityRecognitionMetrics(ProgramMetrics):
    """Activity Recognition program metric class."""

    def __init__(
        self,
        protocol: str,
        video_id: int,
        novel: int,
        detection: int,
        classification: int,
        spatial: int,
        temporal: int,
    ) -> None:
        """
        Initialize.

        Args:
            protocol: Name of the protocol.
            video_id: Column id for video
            novel: Column id for predicting if change was detected
            detection: Column id for predicting sample wise novelty
            classification: Column id for predicting sample wise classes
            spatial: Column id for predicting spatial attribute
            temporal: Column id for predicting temporal attribute

        Returns:
            None
        """
        super().__init__(protocol)
        self.activity_id = video_id
        self.novel_id = novel
        self.detection_id = detection
        self.classification_id = classification
        self.spatial_id = spatial
        self.temporal_id = temporal

    def m_acc(
        self,
        gt_novel: DataFrame,
        p_class: DataFrame,
        gt_class: DataFrame,
        round_size: int,
        asymptotic_start_round: int,
    ) -> Dict:
        """
        m_acc function.

        Args:
            gt_novel: ground truth detections for N videos (Dimension: N X 1)
            p_class: class predictions with video id for N videos (Dimension: N X 90 [vid,novel_class,88 known class])
            gt_class: ground truth classes for N videos (Dimension: N X 1)
            round_size: size of the round
            asymptotic_start_round: asymptotic samples considered for computing metrics

        Returns:
            Dictionary containing top1, top3 accuracy over the test, pre and post novelty.
        """
        class_prob = p_class.iloc[:, range(1, p_class.shape[1])].to_numpy()
        gt_class_idx = gt_class.to_numpy()
        return m_acc(
            gt_novel, class_prob, gt_class_idx, round_size, asymptotic_start_round
        )

    def m_acc_round_wise(
        self, p_class: DataFrame, gt_class: DataFrame, round_id: int
    ) -> Dict:
        """
        m_acc_round_wise function.

        Args:
            p_class: detection predictions
            gt_class: ground truth classes
            round_id: round identifier

        Returns:
            Dictionary containing top1, top3 accuracy for a round
        """
        class_prob = p_class.iloc[:, range(1, p_class.shape[1])].to_numpy()
        gt_class_idx = gt_class.to_numpy()
        top1_acc = topk_accuracy(class_prob, gt_class_idx, k=1)
        top3_acc = topk_accuracy(class_prob, gt_class_idx, k=3)
        return {
            f"top1_accuracy_round_{round_id}": top1_acc,
            f"top3_accuracy_round_{round_id}": top3_acc,
        }

    def m_num(self, p_novel: DataFrame, gt_novel: DataFrame) -> Dict:
        """
        m_num function.

        Args:
            p_novel: detection predictions for N videos (Dimension: N X 1)
            gt_novel: ground truth detections for N videos (Dimension: N X 1)

        Returns:
            Difference between the novelty introduction and predicting change in world.
        """
        return m_num(p_novel, gt_novel)

    def m_num_stats(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        m_num_stats function.

        Args:
            p_novel: detection predictions for N videos (Dimension: N X 1)
            gt_novel: ground truth detections for N videos (Dimension: N X 1)

        Returns:
            Dictionary containing indices for novelty introduction and change in world prediction.
        """
        return m_num_stats(p_novel, gt_novel)

    def m_ndp(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        m_ndp function.

        Args:
            p_novel: detection predictions for N videos (Dimension: N X 1)
            gt_novel: ground truth detections for N videos (Dimension: N X 1)

        Returns:
            Dictionary containing novelty detection performance over the test.
        """
        return m_ndp(p_novel, gt_novel, mode="full_test")

    def m_ndp_pre(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        m_ndp_pre function.

        Args:
            p_novel: detection predictions for N videos (Dimension: N X 1)
            gt_novel: ground truth detections for N videos (Dimension: N X 1)

        Returns:
            Dictionary containing detection performance pre novelty.
        """
        return m_ndp(p_novel, gt_novel, mode="pre_novelty")

    def m_ndp_post(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        m_ndp_post function.

        Args:
            p_novel: detection predictions for N videos (Dimension: N X 1)
            gt_novel: ground truth detections for N videos (Dimension: N X 1)

        Returns:
            Dictionary containing detection performance post novelty.
        """
        return m_ndp(p_novel, gt_novel, mode="post_novelty")

    def m_ndp_failed_reaction(
        self,
        p_novel: DataFrame,
        gt_novel: DataFrame,
        p_class: DataFrame,
        gt_class: DataFrame,
    ) -> Dict:
        """
        m_ndp_failed_reaction function.

        Args:
            p_novel: detection predictions for N videos (Dimension: N X 1)
            gt_novel: ground truth detections for N videos (Dimension: N X 1)
            p_class: class predictions with video id for N videos (Dimension: N X 90 [vid,novel_class,88 known class])
            gt_class: ground truth classes for N videos (Dimension: N X 1)

        Returns:
            Dictionary containing TP, FP, TN, FN, top1, top3 accuracy over the test.
        """
        class_prob = p_class.iloc[:, range(1, p_class.shape[1])].to_numpy()
        gt_class_idx = gt_class.to_numpy()
        return m_ndp_failed_reaction(p_novel, gt_novel, class_prob, gt_class_idx)

    def m_accuracy_on_novel(
        self, p_class: DataFrame, gt_class: DataFrame, gt_novel: DataFrame
    ) -> Dict:
        """
        m_accuracy_on_novel function.

        Args:
            p_class: class predictions with video id for N videos (Dimension: N X 90 [vid,novel_class,88 known class])
            gt_class: ground truth classes for N videos (Dimension: N X 1)
            gt_novel: ground truth detections for N videos (Dimension: N X 1)

        Returns:
            Accuracy on novely samples
        """
        class_prob = p_class.iloc[:, range(1, p_class.shape[1])].to_numpy()
        gt_class_idx = gt_class.to_numpy()
        return m_accuracy_on_novel(class_prob, gt_class_idx, gt_novel)

    def m_is_cdt_and_is_early(self, gt_idx: int, ta2_idx: int, test_len: int) -> Dict:
        """
        m_is_cdt_and_is_early function.

        Args:
            gt_idx: Index when novelty is introduced
            ta2_idx: Index when change is detected
            test_len: Length of test

        Returns
            Dictionary containing boolean showing if change was was detected and if it was detected early
        """
        is_cdt = (ta2_idx >= gt_idx) & (ta2_idx < test_len)
        is_early = ta2_idx < gt_idx
        return {"Is CDT": is_cdt, "Is Early": is_early}

    def m_nrp(self, ta2_acc: Dict, baseline_acc: Dict) -> Dict:
        """
        m_nrp function.

        Args:
            ta2_acc: Accuracy scores for the agent
            baseline_acc: Accuracy scores for baseline

        Returns:
            Reaction performance for the agent
        """
        nrp = {}
        nrp["M_nrp_post_top3"] = 100 * (ta2_acc["post_top3"] / baseline_acc["pre_top3"])
        nrp["M_nrp_post_top1"] = 100 * (ta2_acc["post_top1"] / baseline_acc["pre_top1"])
        return nrp



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sail_on_client',
 'sail_on_client.agent',
 'sail_on_client.configs',
 'sail_on_client.configs.algorithms',
 'sail_on_client.configs.base',
 'sail_on_client.configs.harness',
 'sail_on_client.configs.protocol',
 'sail_on_client.configs.protocol.detection',
 'sail_on_client.configs.protocol.eval',
 'sail_on_client.configs.protocol.feedback',
 'sail_on_client.configs.protocol.resume_session',
 'sail_on_client.configs.protocol.save_attributes',
 'sail_on_client.configs.protocol.save_features',
 'sail_on_client.configs.protocol.use_attributes',
 'sail_on_client.configs.protocol.use_features',
 'sail_on_client.errors',
 'sail_on_client.evaluate',
 'sail_on_client.feedback',
 'sail_on_client.harness',
 'sail_on_client.protocol',
 'sail_on_client.utils']

package_data = \
{'': ['*']}

install_requires = \
['hydra-colorlog>=1.1.0,<2.0.0',
 'hydra-core>=1.1.1,<2.0.0',
 'importlib-metadata>=1.6.0',
 'matplotlib>=3.2.2,<4.0.0',
 'nltk>=3.6.7,<4.0.0',
 'numpy>=1.18.3',
 'pandas>=1.0.3',
 'requests>=2.23.0',
 'scriptconfig==0.5.6',
 'tenacity>=7.0.0,<8.0.0',
 'tinker-engine>=0.10,<0.11',
 'torch>=1.6.0,<2.0.0',
 'types-dataclasses>=0.6.4,<0.7.0',
 'ubelt==0.9.2']

entry_points = \
{'console_scripts': ['sail-on-client = '
                     'sail_on_client.client_launcher:client_launcher'],
 'smqtk_plugins': ['CONDDAProtocol = sail_on_client.protocol.condda_protocol',
                   'LocalHarness = sail_on_client.harness.local_harness',
                   'MockCONDDAAgent = sail_on_client.agent.mock_condda_agents',
                   'MockONDAgent = sail_on_client.agent.mock_ond_agents',
                   'ONDProtocol = sail_on_client.protocol.ond_protocol',
                   'ParHarness = sail_on_client.harness.par_harness',
                   'PreComputedAgent = '
                   'sail_on_client.agent.pre_computed_detector',
                   'PreComputedReactionAgent = '
                   'sail_on_client.agent.pre_computed_reaction_agent']}

setup_kwargs = {
    'name': 'sail-on-client',
    'version': '0.25.0',
    'description': 'Client and Protocols for DARPA sail-on',
    'long_description': '# Sail-On Client and Protocols\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)\n[![CI Workflow](https://github.com/darpa-sail-on/sail-on-client/actions/workflows/ci.yml/badge.svg)](https://gitlab.kitware.com/darpa-sail-on/sail-on-client/-/commits/master)\n[![codecov](https://codecov.io/gh/darpa-sail-on/sail-on-client/branch/master/graph/badge.svg?token=300M5S27NE)](https://codecov.io/gh/darpa-sail-on/sail-on-client)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![PRs Welcome](https://img.shields.io/badge/PRs-welcome-brightgreen.svg?style=flat-square)](https://github.com/darpa-sail-on/sail-on-client)\n\nClient and Protocols for DARPA sail-on\n\n### Protocols present in the repository\n1. [OND](https://drive.google.com/file/d/1W2Ex4-eQl1CrAbv67fAN0OJL8kmRtvt2/view?usp=sharing)\n2. [CONDDA](https://drive.google.com/file/d/1sIDLTgXivaguVfSp3g1qfe7sqiUcvFLA/view?usp=sharing)\n\n## Requirements\n\n1. [Python 3.7](https://www.python.org/downloads/release/python-370/)\n2. [Poetry >= 1.1.0](https://github.com/python-poetry/poetry)\n\n## Installation\n\n1. Install Poetry following the instructions available in the [installation page](https://python-poetry.org/docs/#installation)\n\n2. Clone the repositories associated with different components in a working directory\n    ```\n      git clone https://github.com/darpa-sail-on/sail-on-client.git\n    ```\n   This would create sail-on-client directories in your working directory\n\n\n3. Install the different components in a virtual environment\n   ```\n     cd sail-on-client\n     poetry install\n     poetry shell\n   ```\n\n\n## Running Client and Server with Different Algorithms\n\nNote: If you are using the server setup by PAR, update the `url` to `http://3.32.8.161:5000`\nin sail_on_client/protocol/configuration.json and skip step 1 of running the server.\n\n1. [Instructions for running M6 Algorithms](M6-ALGO.md)\n1. [Instructions for running M12 Algorithms](M12-ALGO.md)\n1. [Instructions for running M18 Algorithms](M18-ALGO.md)\n\n\n## Publishing on PYPI\nsail-on-client uses github actions to publish packages on pypi. The action is triggered when a semver tag is pushed to the repository.\n\nWe support the following version format <major>.<minor>.<patch> and <major>.<minor>.<patch>-alpha.<alpha-version> for tags.\nTo publish a package on pypi, the tag must match with the version maintained in pyproject.toml.\nThis is implemented as a mandatory check in the workflow. Poetry provides support for both querying and bumping version via cli.\nPlease refer to [version](https://python-poetry.org/docs/cli/#version) for more details.\n\nThus to publish sail-on-client on pypi use the following commands\n\n1. Bump the version in pyproject.toml using `poetry version <version_rule>`.\n2. Update `__version__` with the new version in `sail_on_client/__init__.py`.\n3. Use poetry version --short to determine the version that would be used in the tag.\n4. Generate and push the tag using\n   ```\n     git tag <package-version>\n     git push origin --tags\n   ```\n\n## Acknowledgement of Support and Disclaimer\n\nThis material is based upon work supported by the Defense Advanced Research Projects Agency (DARPA) under Contract No. HR001120C0055. Any opinions, findings and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the DARPA.\n',
    'author': 'Ameya Shringi',
    'author_email': 'ameya.shringi@kitware.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Image Classification Feedback."""

from sail_on_client.harness.par_harness import ParHarness
from sail_on_client.harness.local_harness import LocalHarness
from sail_on_client.feedback.feedback import Feedback
from typing import Union

SUPPORTED_FEEDBACK = ["classification", "score"]


class ImageClassificationFeedback(Feedback):
    """Feedback for image classification."""

    def __init__(
        self,
        first_budget: int,
        income_per_batch: int,
        maximum_budget: int,
        interface: Union[LocalHarness, ParHarness],
        session_id: str,
        test_id: str,
        feedback_type: str,
    ) -> None:
        """
        Initialize image classification feedback object.

        Args:
            first_budget: Initial budget
            income_per_batch: Additional labels added after every batch
            maximum_budget: Max labels that can be requested
            interface: An instance of evaluation interface
            session_id: Session identifier
            test_id: Test identifier
            feedback_type: Type of feedback that can be requested

        Returns:
            None
        """
        if feedback_type not in SUPPORTED_FEEDBACK:
            raise ValueError(f"Unsupported feedback_type {feedback_type}")
        super(ImageClassificationFeedback, self).__init__(
            first_budget,
            income_per_batch,
            maximum_budget,
            interface,
            session_id,
            test_id,
            feedback_type,
        )



"""Mocks mainly used for testing protocols."""

from sail_on_client.checkpointer import Checkpointer
from sail_on_client.agent.ond_agent import ONDAgent

from typing import Dict, Any, Tuple, Callable

import logging
import os
import shutil
import torch

log = logging.getLogger(__name__)


class MockONDAgent(ONDAgent):
    """Mock Detector for OND Protocol."""

    def __init__(self) -> None:
        """Construct Mock OND Detector."""
        super().__init__()
        self.step_dict: Dict[str, Callable] = {
            "Initialize": self.initialize,
            "FeatureExtraction": self.feature_extraction,
            "WorldDetection": self.world_detection,
            "NoveltyClassification": self.novelty_classification,
            "NoveltyAdaption": self.novelty_adaptation,
            "NoveltyCharacterization": self.novelty_characterization,
        }

    def initialize(self, toolset: Dict) -> None:
        """
        Algorithm Initialization.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            None
        """
        pass

    def feature_extraction(
        self, toolset: Dict
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Feature extraction step for the algorithm.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            Tuple of dictionary
        """
        self.dataset = toolset["dataset"]
        return {}, {}

    def world_detection(self, toolset: Dict) -> str:
        """
        Detect change in world ( Novelty has been introduced ).

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            path to csv file containing the results for change in world
        """
        dataset_dir = os.path.dirname(self.dataset)
        dst_file = os.path.join(dataset_dir, "wc.csv")
        shutil.copyfile(self.dataset, dst_file)
        return dst_file

    def novelty_classification(self, toolset: Dict) -> str:
        """
        Classify data provided in known classes and unknown class.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            path to csv file containing the results for novelty classification step
        """
        dataset_dir = os.path.dirname(self.dataset)
        dst_file = os.path.join(dataset_dir, "ncl.csv")
        shutil.copyfile(self.dataset, dst_file)
        return dst_file

    def novelty_adaptation(self, toolset: Dict) -> None:
        """
        Update models based on novelty classification and characterization.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            None
        """
        pass

    def novelty_characterization(self, toolset: Dict) -> str:
        """
        Characterize novelty by clustering different novel samples.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            path to csv file containing the results for novelty characterization step
        """
        dataset_dir = os.path.dirname(self.dataset)
        dst_file = os.path.join(dataset_dir, "nc.csv")
        shutil.copyfile(self.dataset, dst_file)
        return dst_file

    def execute(self, toolset: Dict, step_descriptor: str) -> Any:
        """
        Execute method used by the protocol to run different steps.

        Args:
            toolset (dict): Dictionary containing parameters for different steps
            step_descriptor (str): Name of the step
        """
        log.info(f"Executing {step_descriptor}")
        return self.step_dict[step_descriptor](toolset)


class MockONDAgentWithAttributes(MockONDAgent):
    """Mock Detector for testing checkpointing."""

    def __init__(self) -> None:
        """
        Detector constructor.

        Args:
            toolset (dict): Dictionary containing parameters for the constructor
        """
        MockONDAgent.__init__(self)

    def feature_extraction(
        self, toolset: Dict
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Feature extraction step for the algorithm.

        Args:
            toolset (dict): Dictionary containing parameters for different steps

        Return:
            Tuple of dictionary
        """
        self.dummy_dict = toolset["dummy_dict"]
        self.dummy_list = toolset["dummy_list"]
        self.dummy_tuple = toolset["dummy_tuple"]
        self.dummy_tensor = toolset["dummy_tensor"]
        self.dummy_val = toolset["dummy_val"]
        return {}, {}


class MockONDAdapterWithCheckpoint(Checkpointer, MockONDAgent):
    """Mock Adapter for testing checkpointing."""

    def __init__(self, toolset: Dict) -> None:
        """
        Detector constructor.

        Args:
            toolset (dict): Dictionary containing parameters for the constructor
        """
        MockONDAgent.__init__(self)
        Checkpointer.__init__(self, toolset)
        self.detector = MockONDAgentWithAttributes()

    def get_config(self) -> Dict:
        """
        Get config for the plugin.

        Returns:
            Parameters for the agent
        """
        config = super().get_config()
        config.update(self.toolset)
        return config

    def execute(self, toolset: Dict, step_descriptor: str) -> Any:
        """
        Execute method used by the protocol to run different steps.

        Args:
            toolset (dict): Dictionary containing parameters for different steps
            step_descriptor (str): Name of the step
        """
        log.info(f"Executing {step_descriptor}")
        return self.detector.step_dict[step_descriptor](toolset)

    def __eq__(self, other: object) -> bool:
        """
        Overriden method to compare two mock adapters.

        Args:
            other (MockONDAdapterWithCheckpoint): Another instance of mock adapter

        Return:
            True if both instances have same attributes
        """
        if not isinstance(other, MockONDAdapterWithCheckpoint):
            return NotImplemented

        return (
            self.detector.dummy_dict == other.detector.dummy_dict
            and self.detector.dummy_list == other.detector.dummy_list
            and self.detector.dummy_tuple == other.detector.dummy_tuple
            and bool(
                torch.all(
                    torch.eq(self.detector.dummy_tensor, other.detector.dummy_tensor)
                )
            )
            and self.detector.dummy_val == other.detector.dummy_val
        )



"""CLI For client."""
import hydra
import logging
import os
from omegaconf import DictConfig, OmegaConf
from omegaconf.errors import MissingMandatoryValue
from tinker.configuration import process_config
from typing import cast

from sail_on_client.protocol.visual_protocol import VisualProtocol

log = logging.getLogger(__name__)


@hydra.main(config_path="configs", config_name="default")
def client_launcher(cfg: DictConfig) -> None:
    """
    Primary entrypoint for the client.

    Args:
        cfg: config dictionary obtained from configs provided on terminal

    Returns:
        None
    """
    log.info(cfg)
    # Hack to set save directory based on working directory
    try:
        cfg_dict = OmegaConf.to_container(
            cfg.protocol, resolve=True, throw_on_missing=True
        )
    except MissingMandatoryValue:
        # Set save directory to output directory that hydra is using
        cfg.protocol.smqtk.config.save_dir = os.getcwd()
    finally:
        cfg_dict = OmegaConf.to_container(
            cfg.protocol, resolve=True, throw_on_missing=True
        )
    processed_config = process_config(cfg_dict)  # type: ignore
    if issubclass(type(processed_config), VisualProtocol):
        protocol = cast(VisualProtocol, processed_config)
        protocol.run_protocol({})
    else:
        raise NotImplementedError(f"Failed to create a protocol from {cfg}")


if __name__ == "__main__":
    client_launcher()



"""Dataclasses for the protocols."""

from dataclasses import dataclass
import logging
from typing import Dict, List, Any, Union
from sail_on_client.utils.utils import merge_dictionaries
from sail_on_client.feedback.image_classification_feedback import (
    ImageClassificationFeedback,
)
from sail_on_client.feedback.document_transcription_feedback import (
    DocumentTranscriptionFeedback,
)
from sail_on_client.feedback.activity_recognition_feedback import (
    ActivityRecognitionFeedback,
)

try:
    from importlib.metadata import version, PackageNotFoundError  # type:ignore
except ModuleNotFoundError:
    from importlib_metadata import version, PackageNotFoundError

log = logging.getLogger(__name__)


@dataclass
class AlgorithmAttributes:
    """Class for storing attributes of algorithm present in the protocol."""

    name: str
    detection_threshold: float
    instance: Any
    is_baseline: bool
    is_reaction_baseline: bool
    package_name: str
    parameters: Dict
    session_id: str
    test_ids: List[str]

    def named_version(self) -> str:
        """
        Compute version of an algorithm.

        Returns:
            A string containing name and the version number
        """
        try:
            if self.package_name:
                version_number = version(self.package_name)
            else:
                log.warn("No package_name provided. Using 0.0.1 as stand in.")
                version_number = "0.0.1"
        except PackageNotFoundError:
            log.warn(
                "Failed to detect the version of the algorithm. Using 0.0.1 as stand in."
            )
            version_number = "0.0.1"
        return f"{self.name}-{version_number}"

    def remove_completed_tests(self, finished_tests: List[str]) -> None:
        """
        Remove finished tests from test_ids.

        Args:
            finished_tests: List of tests that are complete

        Returns:
            None
        """
        test_set = set(self.test_ids)
        ftest_set = set(finished_tests)
        self.test_ids = list(test_set ^ ftest_set)

    def merge_detector_params(
        self, detector_params: Dict, exclude_keys: List = None
    ) -> None:
        """
        Merge common parameters with algorithm specific parameters with exclusions.

        Args:
            detector_params: Dictionary of common parameters
            exclude_keys: List of keys that should be excluded in the merge

        Returns:
            None
        """
        if not exclude_keys:
            exclude_keys = []
        self.parameters = merge_dictionaries(
            self.parameters, detector_params, exclude_keys
        )


@dataclass
class InitializeParams:
    """Class for storing parameters that are used to initialize the algorithm."""

    parameters: Dict
    session_id: str
    test_id: str
    feedback_instance: Union[
        ImageClassificationFeedback,
        DocumentTranscriptionFeedback,
        ActivityRecognitionFeedback,
        None,
    ]

    def get_toolset(self) -> Dict:
        """
        Convert the data present in the class into a dictionary.

        Returns
            A dictionary with data associated with the class
        """
        algorithm_toolset = self.parameters.copy()
        algorithm_toolset["session_id"] = self.session_id
        algorithm_toolset["test_id"] = self.test_id
        algorithm_toolset["test_type"] = ""
        if self.feedback_instance:
            algorithm_toolset["FeedbackInstance"] = self.feedback_instance
        return algorithm_toolset


@dataclass
class NoveltyClassificationParams:
    """Class for storing parameters associated novelty classification in an algorithm."""

    features_dict: Dict
    logit_dict: Dict
    round_id: int

    def get_toolset(self) -> Dict:
        """
        Convert the data present in the class into a dictionary.

        Returns
            A dictionary with data associated with the class
        """
        return {
            "features_dict": self.features_dict,
            "logit_dict": self.logit_dict,
            "round_id": self.round_id,
        }


@dataclass
class NoveltyAdaptationParams:
    """Class for storing parameters associated novelty adaptation with an algorithm."""

    round_id: int

    def get_toolset(self) -> Dict:
        """
        Convert the data present in the class into a dictionary.

        Returns
            A dictionary with data associated with the class
        """
        return {
            "round_id": self.round_id,
        }


@dataclass
class NoveltyCharacterizationParams:
    """Class for storing parameters associated novelty characterization with an algorithm."""

    dataset_ids: List[str]

    def get_toolset(self) -> Dict:
        """
        Convert the data present in the class into a dictionary.

        Returns
            A dictionary with data associated with the class
        """
        return {"dataset_ids": self.dataset_ids}
